import pybullet as p
import time
import pybullet_data

p.connect(p.GUI)
p.setAdditionalSearchPath(pybullet_data.getDataPath())
p.setPhysicsEngineParameter(allowedCcdPenetration=0.0)

terrain_mass = 0
terrain_visual_shape_id = -1
terrain_position = [0, 0, 0]
terrain_orientation = [0, 0, 0, 1]
terrain_collision_shape_id = p.createCollisionShape(shapeType=p.GEOM_MESH,
                                                    fileName="terrain.obj",
                                                    flags=p.GEOM_FORCE_CONCAVE_TRIMESH |
                                                    p.GEOM_CONCAVE_INTERNAL_EDGE,
                                                    meshScale=[0.5, 0.5, 0.5])
p.createMultiBody(terrain_mass, terrain_collision_shape_id, terrain_visual_shape_id,
                  terrain_position, terrain_orientation)

useMaximalCoordinates = True
sphereRadius = 0.005
colSphereId = p.createCollisionShape(p.GEOM_SPHERE, radius=sphereRadius)
colBoxId = p.createCollisionShape(p.GEOM_BOX,
                                  halfExtents=[sphereRadius, sphereRadius, sphereRadius])

mass = 1
visualShapeId = -1

for i in range(5):
  for j in range(5):
    for k in range(5):
      #if (k&2):
      sphereUid = p.createMultiBody(
          mass,
          colSphereId,
          visualShapeId, [-i * 5 * sphereRadius, j * 5 * sphereRadius, k * 2 * sphereRadius + 1],
          useMaximalCoordinates=useMaximalCoordinates)
      #else:
      #	sphereUid = p.createMultiBody(mass,colBoxId,visualShapeId,[-i*2*sphereRadius,j*2*sphereRadius,k*2*sphereRadius+1], useMaximalCoordinates=useMaximalCoordinates)
      p.changeDynamics(sphereUid,
                       -1,
                       spinningFriction=0.001,
                       rollingFriction=0.001,
                       linearDamping=0.0)
      p.changeDynamics(sphereUid, -1, ccdSweptSphereRadius=0.002)

p.setGravity(0, 0, -10)

pts = p.getContactPoints()
print("num points=", len(pts))
print(pts)
while (p.isConnected()):
  time.sleep(1. / 240.)
  p.stepSimulation()



from pybullet_utils import bullet_client as bc
from pybullet_utils import urdfEditor as ed
import pybullet
import pybullet_data
import time

p0 = bc.BulletClient(connection_mode=pybullet.DIRECT)
p0.setAdditionalSearchPath(pybullet_data.getDataPath())

p1 = bc.BulletClient(connection_mode=pybullet.DIRECT)
p1.setAdditionalSearchPath(pybullet_data.getDataPath())

#can also connect using different modes, GUI, SHARED_MEMORY, TCP, UDP, SHARED_MEMORY_SERVER, GUI_SERVER

husky = p1.loadURDF("husky/husky.urdf", flags=p0.URDF_USE_IMPLICIT_CYLINDER)
kuka = p0.loadURDF("kuka_iiwa/model.urdf")

ed0 = ed.UrdfEditor()
ed0.initializeFromBulletBody(husky, p1._client)
ed1 = ed.UrdfEditor()
ed1.initializeFromBulletBody(kuka, p0._client)
#ed1.saveUrdf("combined.urdf")

parentLinkIndex = 0

jointPivotXYZInParent = [0, 0, 0]
jointPivotRPYInParent = [0, 0, 0]

jointPivotXYZInChild = [0, 0, 0]
jointPivotRPYInChild = [0, 0, 0]

newjoint = ed0.joinUrdf(ed1, parentLinkIndex, jointPivotXYZInParent, jointPivotRPYInParent,
                        jointPivotXYZInChild, jointPivotRPYInChild, p0._client, p1._client)
newjoint.joint_type = p0.JOINT_FIXED

ed0.saveUrdf("combined.urdf")

print(p0._client)
print(p1._client)
print("p0.getNumBodies()=", p0.getNumBodies())
print("p1.getNumBodies()=", p1.getNumBodies())

pgui = bc.BulletClient(connection_mode=pybullet.GUI)
pgui.configureDebugVisualizer(pgui.COV_ENABLE_RENDERING, 0)

orn = [0, 0, 0, 1]
ed0.createMultiBody([0, 0, 0], orn, pgui._client)
pgui.setRealTimeSimulation(1)

pgui.configureDebugVisualizer(pgui.COV_ENABLE_RENDERING, 1)

while (pgui.isConnected()):
  pgui.getCameraImage(320, 200, renderer=pgui.ER_BULLET_HARDWARE_OPENGL)
  time.sleep(1. / 240.)



# Lint as: python3
"""A static gait controller for a quadruped robot. Experimental code."""

import gin
import numpy as np
from pybullet_envs.minitaur.agents.baseline_controller import foot_stepper

toe_pos_local_ref = np.array([[0.1478, -0.11459, -0.45576],
                              [0.1478, 0.11688, -0.45576],
                              [-0.2895, -0.11459, -0.45576],
                              [-0.2895, 0.11688, -0.45576]])


@gin.configurable
class StaticGaitController(object):
  """A static gait controller for a quadruped robot."""

  def __init__(self, robot):
    self._robot = robot
    self._toe_ids = tuple(robot.urdf_loader.get_end_effector_id_dict().values())
    self._wait_count = 0
    self._stepper = foot_stepper.FootStepper(self._robot.pybullet_client,
                                             self._toe_ids, toe_pos_local_ref)

  def act(self, observation):
    """Computes actions based on observations."""
    del observation
    p = self._robot.pybullet_client
    quadruped = self._robot.robot_id
    step_input = foot_stepper.StepInput()
    ls = p.getLinkStates(
        quadruped, self._toe_ids, computeForwardKinematics=True)
    toe_pos_world = np.array([ls[0][0], ls[1][0], ls[2][0], ls[3][0]])
    base_com_pos, base_com_orn = p.getBasePositionAndOrientation(quadruped)
    new_pos_world = np.array([0, 0, 0])

    if self._stepper.is_com_stable() and not self._stepper.move_swing_foot:
      self._wait_count += 1
      if self._wait_count == 20:
        self._stepper.next_foot()
      if self._wait_count > 50:
        self._wait_count = 0
        step_dist = 0.15
        print("time {}, make a step of {}".format(
            self._robot.GetTimeSinceReset(), step_dist))
        new_pos_local = self._stepper.get_reference_pos_swing_foot()
        new_pos_local[0] += step_dist
        new_pos_world, _ = p.multiplyTransforms(base_com_pos, base_com_orn,
                                                new_pos_local, [0, 0, 0, 1])
        self._stepper.swing_foot()

    step_input.new_pos_world = new_pos_world
    step_input.base_com_pos = base_com_pos
    step_input.base_com_orn = base_com_orn
    step_input.toe_pos_world = toe_pos_world
    step_input.dt = 1.0 / 250
    step_output = self._stepper.update(step_input)

    # Finds joint poses to achieve toePosWorld
    desired_joint_angles = self._robot.motor_angles_from_foot_positions(
        foot_positions=step_output.new_toe_pos_world,
        position_in_world_frame=True)[1]
    return desired_joint_angles



import json
import math
import random
import numpy as np
from os import listdir
from os.path import isfile, join

from scipy.spatial.transform import Rotation as R
from scipy.spatial.transform import Slerp


class MotionCaptureDataMultiClip(object):

    def __init__(self):
        self.Reset()

    def Reset(self):
        self._motion_data = {}
        self._num_frames_min = 9999
        self._num_frames_max = -1
        self._num_clips = -1
        self._names = []

    def Load(self, path):
        if isfile(path):
            files = [path]
        else:
            files = [join(path, f) for f in listdir(path) if isfile(join(path, f))]
            files.sort()
        for i in range(len(files)):
            with open(files[i], 'r') as f:
                self._names.append(files[i].split('/')[-1])
                self._motion_data[i] = json.load(f)
                t = len(self._motion_data[i]['Frames'])
                self._num_frames_min = min(t, self._num_frames_min)
                self._num_frames_max = max(t, self._num_frames_max)
        self._num_clips = len(self._motion_data)
        self._downsample = True
        if self._downsample:
            self.downsampleClips()
            self._num_frames_max = self._num_frames_min
        else:
            self.upsampleClips()
            self._num_frames_min = self._num_frames_max

    def getNumFrames(self):
        if self._downsample:
            return self._num_frames_min
        else:
            return self._num_frames_max

    def getKeyFrameDuration(self, id=0):
        return self._motion_data[id]['Frames'][0][0]

    def getCycleTime(self):
        keyFrameDuration = self.getKeyFrameDuration()
        cycleTime = keyFrameDuration * (self.getNumFrames() - 1)
        return cycleTime

    def calcCycleCount(self, simTime, cycleTime):
        phases = simTime / cycleTime
        count = math.floor(phases)
        return count

    def computeCycleOffset(self, id=0):
        lastFrame = self.getNumFrames() - 1
        frameData = self._motion_data[id]['Frames'][0]
        frameDataNext = self._motion_data[id]['Frames'][lastFrame]

        basePosStart = [frameData[1], frameData[2], frameData[3]]
        basePosEnd = [frameDataNext[1], frameDataNext[2], frameDataNext[3]]
        self._cycleOffset = [
            basePosEnd[0] - basePosStart[0], basePosEnd[1] - basePosStart[1],
            basePosEnd[2] - basePosStart[2]
        ]
        return self._cycleOffset

    def getNumClips(self):
        return self._num_clips

    def downsampleClips(self):
        for i in range(self._num_clips):
            n_frames = len(self._motion_data[i]['Frames'])
            if n_frames != self._num_frames_min:
                sample = random.sample(range(n_frames), self._num_frames_min)
                sample.sort()
                downsampled = np.array(self._motion_data[i]['Frames'])[sample]
                self._motion_data[i]['Frames'] = downsampled.tolist()
                #s = json.dumps(self._motion_data[i])
                #with open("output/{}".format(self._names[i]), 'w') as f:
                #    f.writelines(s)

    def upsampleClips(self):
        print("Max number of frames: ", self._num_frames_max)
        for i in range(self._num_clips):
            #print("Uspsampling clip number: ", i)
            keyframe_duration = self.getKeyFrameDuration(i)
            old_times = np.arange(0, len(self._motion_data[i]['Frames']) * keyframe_duration, keyframe_duration)
            while len(old_times) < self._num_frames_max:
                new_times, new_vals = self.slerpSingleClip(self._motion_data[i]['Frames'], old_times)
                #print("Number of final frames: ", len(new_vals))
                self._motion_data[i]['Frames'] = new_vals
                old_times = new_times
            #s = json.dumps(self._motion_data[i])
            #with open("output/{}".format(self._names[i]), 'w') as f:
            #    f.writelines(s)


    def slerpSingleClip(self, clip, key_times):
        #print("Number of initial frames: ", len(key_times))
        org_clip = self.quatlist_to_quatlists(clip)
        org_clip = np.asarray(org_clip)
        t = org_clip[:, 0]
        root_pos = org_clip[:, 1]
        key_rots = org_clip[:, 2]
        n_frames = len(key_rots)
        assert len(key_times) == n_frames
        needed_frames = self._num_frames_max - n_frames
        #print("Needed frames: ", needed_frames)
        inter_times = self.calc_inter_times(key_times)
        inter_times = sorted(random.sample(inter_times, min(len(inter_times), needed_frames)))
        #print("Number of frames to interpolate: ", len(inter_times))
        #print("Number of rots: ", len(key_rots[0]))
        inter_joint = []
        for i in range(len(key_rots[0])):
            quats = [rot[i] for rot in key_rots]
            if len(quats[0]) == 4:
                joint = R.from_quat(quats)
                slerp = Slerp(key_times, joint)
                interp_rots = slerp(inter_times)
                interp_rots = interp_rots.as_quat().tolist()
            else:
                interp_rots = []
                for tim in range(len(inter_times)):
                    lb = key_times.tolist().index(max([st for st in key_times if st < inter_times[tim]]))
                    ub = lb + 1
                    #print(lb, ub)
                    new_rot = (quats[ub][0] + quats[lb][0])/2
                    interp_rots.append([new_rot])
            inter_joint.append(interp_rots)
        inter_joint = np.array(inter_joint).T
        #print("Shape of interpolated joints: ", inter_joint.shape)
        old_dict = dict(zip(key_times, key_rots))
        new_dict = dict(zip(inter_times, inter_joint))
        old_root_pos = dict(zip(key_times, root_pos))
        inter_root_pos = self.calc_inter_root_pos(root_pos, key_times, inter_times)
        new_root_pos = dict(zip(inter_times, inter_root_pos))
        new_dict = {**old_dict, **new_dict}
        new_rp_dict = {**old_root_pos, **new_root_pos}
        ord_keys = sorted(new_dict.keys())
        ord_rots = [new_dict[k] for k in ord_keys]
        ord_root_pos = [new_rp_dict[k] for k in ord_keys]
        new_clip = self.quatlists_to_quatlist(t, ord_root_pos, ord_rots)

        return np.array(ord_keys), new_clip

    def quatlist_to_quatlists(self, clip):
        new_clips = []
        for c in clip:
            t = c[0]
            root_pos = c[1:4]
            root_rotation = c[4:8]
            chest_rotation = c[8:12]
            neck_rotation = c[12:16]
            right_hip_rotation = c[16:20]
            right_knee_rotation = c[20]
            right_ankle_rotation = c[21:25]
            right_shoulder_rotation = c[25:29]
            right_elbow_rotation = c[29]
            left_hip_rotation = c[30:34]
            left_knee_rotation = c[34]
            left_ankle_rotation = c[35:39]
            left_shoulder_rotation = c[39:43]
            left_elbow_rotation = c[43]
            d = [
                t,
                root_pos,
                [
                    self.deepmimic_to_scipy_quaternion(root_rotation),
                    self.deepmimic_to_scipy_quaternion(chest_rotation),
                    self.deepmimic_to_scipy_quaternion(neck_rotation),
                    self.deepmimic_to_scipy_quaternion(right_hip_rotation),
                    [right_knee_rotation],
                    self.deepmimic_to_scipy_quaternion(right_ankle_rotation),
                    self.deepmimic_to_scipy_quaternion(right_shoulder_rotation),
                    [right_elbow_rotation],
                    self.deepmimic_to_scipy_quaternion(left_hip_rotation),
                    [left_knee_rotation],
                    self.deepmimic_to_scipy_quaternion(left_ankle_rotation),
                    self.deepmimic_to_scipy_quaternion(left_shoulder_rotation),
                    [left_elbow_rotation]
                ]
            ]
            new_clips.append(d)
        return new_clips

    def deepmimic_to_scipy_quaternion(self, quat):
        return quat[1:] + [quat[0]]

    def scipy_to_deepmimic_quaternion(self, quat):
        return [quat[-1]] + quat[:-1]

    def calc_inter_times(self, times, method="intermediate"):
        if method == "intermediate":
            inter_times = []
            for i in range(1, len(times)):
                it = (times[i] - times[i-1])/2 + times[i-1]
                inter_times.append(it)
            return inter_times

    def calc_inter_root_pos(self, root_pos, times, inter_times):
        inter_root_pos = []
        all_times = sorted([*times.tolist(), *inter_times])
        for i in range(len(inter_times)):
            low_index = times.tolist().index(all_times[all_times.index(inter_times[i]) - 1])
            up_index = times.tolist().index(all_times[all_times.index(inter_times[i]) + 1])
            assert low_index == up_index - 1
            inter_root_pos.append(((np.array(root_pos[up_index]) + np.array(root_pos[low_index]))/2).tolist())
        return inter_root_pos

    def quatlists_to_quatlist(self, t, ord_root_pos, ord_rots):
        delta_t = t[0]
        new_quats = self.merge_quaternions(ord_rots)
        a = []
        for i in range(len(ord_root_pos)):
            rot = new_quats[i]

            a.append([
                delta_t,
                *ord_root_pos[i],
                *rot
            ])
        return a

    def merge_quaternions(self, rotations):
        quats = []
        for rot in rotations:
            rots = []
            for r in rot:
                if len(r) == 4:
                    r = self.scipy_to_deepmimic_quaternion(r)
                rots += [el for el in r]
            quats.append(rots)
        return quats


# Lint as: python3
"""Quadruped toe position sensor."""

from typing import Any, Callable, Sequence, Text, Tuple, Type, Union

import gin
import gym
import numpy as np

from pybullet_envs.minitaur.envs_v2.sensors import sensor
from pybullet_envs.minitaur.envs_v2.utilities import noise_generators


def _convert_to_np_array(inputs: Union[float, Tuple[float], np.ndarray], dim):
  """Converts the inputs to a numpy array.

  Args:
    inputs: The input scalar or array.
    dim: The dimension of the converted numpy array.

  Returns:
    The converted numpy array.

  Raises:
    ValueError: If the inputs is an array whose dimension does not match the
    provided dimension.
  """
  outputs = None
  if isinstance(inputs, (tuple, np.ndarray)):
    outputs = np.array(inputs)
  else:
    outputs = np.full(dim, inputs)

  if len(outputs) != dim:
    raise ValueError("The inputs array has a different dimension {}"
                     " than provided, which is {}.".format(len(outputs), dim))

  return outputs


@gin.configurable
class ToePositionSensor(sensor.Sensor):
  """A sensor that outputs the toe positions of attached robots or objects."""

  def __init__(
      self,
      name: Text = "toe_position",
      dtype: Type[Any] = np.float64,
      lower_bound: Union[float, Sequence[float]] = -1.0,
      upper_bound: Union[float, Sequence[float]] = 1.0,
      noise_generator: Union[Callable[..., Any],
                             noise_generators.NoiseGenerator] = None,
      sensor_latency: Union[float, Sequence[float]] = 0.0,
  ):
    """Constructor.

    Args:
      name: Name of the sensor.
      dtype: Data type of sensor value.
      lower_bound: The optional lower bounds of the sensor reading.
      upper_bound: The optional upper bounds of the sensor reading.
      noise_generator: Used to add noise to the readings.
      sensor_latency: Single or multiple latency in seconds. See sensor.Sensor
        docstring for details.
    """
    super().__init__(
        name=name,
        sensor_latency=sensor_latency,
        interpolator_fn=sensor.linear_obs_blender)
    self._dtype = dtype
    self._lower_bound = lower_bound
    self._upper_bound = upper_bound
    self._noise_generator = noise_generator

  def set_robot(self, robot):
    self._robot = robot
    num_legs = len(robot.urdf_loader.get_end_effector_id_dict().values())
    lower_bound = _convert_to_np_array(self._lower_bound, num_legs * 3)

    upper_bound = _convert_to_np_array(self._upper_bound, num_legs * 3)

    self._observation_space = self._stack_space(
        gym.spaces.Box(low=lower_bound, high=upper_bound, dtype=self._dtype))

  def _get_original_observation(self) -> Tuple[float, np.ndarray]:
    """Returns raw observation with timestamp."""
    toe_position = np.array(
        self._robot.foot_positions(), dtype=self._dtype).flatten()

    return self._robot.timestamp, toe_position

  def get_observation(self) -> np.ndarray:
    delayed_observation = super().get_observation()
    if self._noise_generator:
      if callable(self._noise_generator):
        return self._noise_generator(delayed_observation)
      else:
        return self._noise_generator.add_noise(delayed_observation)

    return delayed_observation



"""Tests for imagery_utils."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import struct
import numpy as np

from pybullet_envs.minitaur.vision import imagery_pb2
from pybullet_envs.minitaur.vision import imagery_utils
from google3.testing.pybase import googletest


class ImageryUtilsTest(googletest.TestCase):

  def test_convert_bgra_images(self):
    image = imagery_pb2.Image(
        height_px=2,
        width_px=2,
        image_format=imagery_pb2.Image.IMAGE_FORMAT_BGRA_HWC_8U,
        content=b'ABCDABCDABCDABCD',
    )

    image_array = imagery_utils.convert_image_to_array(image)

    self.assertEqual(image_array.dtype, np.uint8)
    self.assertEqual(image_array.shape, (image.height_px, image.width_px, 4))
    self.assertEqual(image_array[0, 0, 0], ord('A'))
    self.assertEqual(image_array[1, 0, 3], ord('D'))

  def test_convert_rgb_images(self):
    image = imagery_pb2.Image(
        height_px=2,
        width_px=2,
        image_format=imagery_pb2.Image.IMAGE_FORMAT_RGB_HWC_8U,
        content=b'ABCABCABCABC',
    )

    image_array = imagery_utils.convert_image_to_array(image)

    self.assertEqual(image_array.dtype, np.uint8)
    self.assertEqual(image_array.shape, (image.height_px, image.width_px, 3))
    self.assertEqual(image_array[0, 0, 0], ord('A'))
    self.assertEqual(image_array[1, 1, 2], ord('C'))

  def test_convert_gray_32bit_images(self):
    image = imagery_pb2.Image(
        height_px=2,
        width_px=3,
        image_format=imagery_pb2.Image.IMAGE_FORMAT_GRAY_HW_32F,
        content=b'AAAABBBBCCCCAAAABBBBCCCC',
    )

    image_array = imagery_utils.convert_image_to_array(image)

    self.assertEqual(image_array.dtype, np.float32)
    self.assertEqual(image_array.shape, (image.height_px, image.width_px))
    self.assertEqual(image_array[0, 2], struct.unpack(b'<f', b'CCCC'))
    self.assertEqual(image_array[1, 1], struct.unpack(b'<f', b'BBBB'))

  def test_convert_gray_16bit_images(self):
    image = imagery_pb2.Image(
        height_px=3,
        width_px=2,
        image_format=imagery_pb2.Image.IMAGE_FORMAT_GRAY_HW_16U,
        content=b'AABBCCAABBCC',
    )

    image_array = imagery_utils.convert_image_to_array(image)

    self.assertEqual(image_array.dtype, np.uint16)
    self.assertEqual(image_array.shape, (image.height_px, image.width_px))
    self.assertEqual(image_array[0, 1], struct.unpack(b'<H', b'BB'))
    self.assertEqual(image_array[2, 1], struct.unpack(b'<H', b'CC'))

  def test_unspecified_image_format(self):
    image = imagery_pb2.Image()
    with self.assertRaises(ValueError):
      imagery_utils.convert_image_to_array(image)


if __name__ == '__main__':
  googletest.main()



import numpy as np


class PDControllerExplicitMultiDof(object):

  def __init__(self, pb):
    self._pb = pb

  def computePD(self, bodyUniqueId, jointIndices, desiredPositions, desiredVelocities, kps, kds,
                maxForces, timeStep):

    numJoints = len(jointIndices)  #self._pb.getNumJoints(bodyUniqueId)
    curPos, curOrn = self._pb.getBasePositionAndOrientation(bodyUniqueId)
    q1 = [curPos[0], curPos[1], curPos[2], curOrn[0], curOrn[1], curOrn[2], curOrn[3]]
    baseLinVel, baseAngVel = self._pb.getBaseVelocity(bodyUniqueId)
    qdot1 = [
        baseLinVel[0], baseLinVel[1], baseLinVel[2], baseAngVel[0], baseAngVel[1], baseAngVel[2], 0
    ]
    qError = [0, 0, 0, 0, 0, 0, 0]
    qIndex = 7
    qdotIndex = 7
    zeroAccelerations = [0, 0, 0, 0, 0, 0, 0]
    for i in range(numJoints):
      js = self._pb.getJointStateMultiDof(bodyUniqueId, jointIndices[i])

      jointPos = js[0]
      jointVel = js[1]
      q1 += jointPos

      if len(js[0]) == 1:
        desiredPos = desiredPositions[qIndex]

        qdiff = desiredPos - jointPos[0]
        qError.append(qdiff)
        zeroAccelerations.append(0.)
        qdot1 += jointVel
        qIndex += 1
        qdotIndex += 1
      if len(js[0]) == 4:
        desiredPos = [
            desiredPositions[qIndex], desiredPositions[qIndex + 1], desiredPositions[qIndex + 2],
            desiredPositions[qIndex + 3]
        ]
        axis = self._pb.getAxisDifferenceQuaternion(desiredPos, jointPos)
        jointVelNew = [jointVel[0], jointVel[1], jointVel[2], 0]
        qdot1 += jointVelNew
        qError.append(axis[0])
        qError.append(axis[1])
        qError.append(axis[2])
        qError.append(0)
        desiredVel = [
            desiredVelocities[qdotIndex], desiredVelocities[qdotIndex + 1],
            desiredVelocities[qdotIndex + 2]
        ]
        zeroAccelerations += [0., 0., 0., 0.]
        qIndex += 4
        qdotIndex += 4

    q = np.array(q1)
    qdot = np.array(qdot1)
    qdotdesired = np.array(desiredVelocities)
    qdoterr = qdotdesired - qdot
    Kp = np.diagflat(kps)
    Kd = np.diagflat(kds)
    p = Kp.dot(qError)
    d = Kd.dot(qdoterr)
    forces = p + d
    maxF = np.array(maxForces)
    forces = np.clip(forces, -maxF, maxF)
    return forces


class PDControllerExplicit(object):

  def __init__(self, pb):
    self._pb = pb

  def computePD(self, bodyUniqueId, jointIndices, desiredPositions, desiredVelocities, kps, kds,
                maxForces, timeStep):
    numJoints = self._pb.getNumJoints(bodyUniqueId)
    jointStates = self._pb.getJointStates(bodyUniqueId, jointIndices)
    q1 = []
    qdot1 = []
    for i in range(numJoints):
      q1.append(jointStates[i][0])
      qdot1.append(jointStates[i][1])
    q = np.array(q1)
    qdot = np.array(qdot1)
    qdes = np.array(desiredPositions)
    qdotdes = np.array(desiredVelocities)
    qError = qdes - q
    qdotError = qdotdes - qdot
    Kp = np.diagflat(kps)
    Kd = np.diagflat(kds)
    forces = Kp.dot(qError) + Kd.dot(qdotError)
    maxF = np.array(maxForces)
    forces = np.clip(forces, -maxF, maxF)
    return forces



# Copyright 2017 The TensorFlow Agents Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Combine multiple environments to step them in batch."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import numpy as np


class BatchEnv(object):
  """Combine multiple environments to step them in batch."""

  def __init__(self, envs, blocking):
    """Combine multiple environments to step them in batch.

    To step environments in parallel, environments must support a
    `blocking=False` argument to their step and reset functions that makes them
    return callables instead to receive the result at a later time.

    Args:
      envs: List of environments.
      blocking: Step environments after another rather than in parallel.

    Raises:
      ValueError: Environments have different observation or action spaces.
    """
    self._envs = envs
    self._blocking = blocking
    observ_space = self._envs[0].observation_space
    if not all(env.observation_space == observ_space for env in self._envs):
      raise ValueError('All environments must use the same observation space.')
    action_space = self._envs[0].action_space
    if not all(env.action_space == action_space for env in self._envs):
      raise ValueError('All environments must use the same observation space.')

  def __len__(self):
    """Number of combined environments."""
    return len(self._envs)

  def __getitem__(self, index):
    """Access an underlying environment by index."""
    return self._envs[index]

  def __getattr__(self, name):
    """Forward unimplemented attributes to one of the original environments.

    Args:
      name: Attribute that was accessed.

    Returns:
      Value behind the attribute name one of the wrapped environments.
    """
    return getattr(self._envs[0], name)

  def step(self, actions):
    """Forward a batch of actions to the wrapped environments.

    Args:
      actions: Batched action to apply to the environment.

    Raises:
      ValueError: Invalid actions.

    Returns:
      Batch of observations, rewards, and done flags.
    """
    for index, (env, action) in enumerate(zip(self._envs, actions)):
      if not env.action_space.contains(action):
        message = 'Invalid action at index {}: {}'
        raise ValueError(message.format(index, action))
    if self._blocking:
      transitions = [env.step(action) for env, action in zip(self._envs, actions)]
    else:
      transitions = [env.step(action, blocking=False) for env, action in zip(self._envs, actions)]
      transitions = [transition() for transition in transitions]
    observs, rewards, dones, infos = zip(*transitions)
    observ = np.stack(observs)
    reward = np.stack(rewards)
    done = np.stack(dones)
    info = tuple(infos)
    return observ, reward, done, info

  def reset(self, indices=None):
    """Reset the environment and convert the resulting observation.

    Args:
      indices: The batch indices of environments to reset; defaults to all.

    Returns:
      Batch of observations.
    """
    if indices is None:
      indices = np.arange(len(self._envs))
    if self._blocking:
      observs = [self._envs[index].reset() for index in indices]
    else:
      observs = [self._envs[index].reset(blocking=False) for index in indices]
      observs = [observ() for observ in observs]
    observ = np.stack(observs)
    return observ

  def close(self):
    """Send close messages to the external process and join them."""
    for env in self._envs:
      if hasattr(env, 'close'):
        env.close()



#add parent dir to find package. Only needed for source code build, pip install doesn't need it.
import os, inspect
currentdir = os.path.dirname(os.path.abspath(inspect.getfile(inspect.currentframe())))
parentdir = os.path.dirname(os.path.dirname(currentdir))
os.sys.path.insert(0, parentdir)

import gym
import time

from baselines import deepq
from pybullet_envs.bullet.cartpole_bullet import CartPoleBulletEnv


def main():
  env = gym.make('CartPoleBulletEnv-v1')
  act = deepq.load("cartpole_model.pkl")

  while True:
    obs, done = env.reset(), False
    print("obs")
    print(obs)
    print("type(obs)")
    print(type(obs))
    episode_rew = 0
    while not done:
      env.render()

      o = obs[None]
      aa = act(o)
      a = aa[0]
      obs, rew, done, _ = env.step(a)
      episode_rew += rew
      time.sleep(1. / 240.)
    print("Episode reward", episode_rew)


if __name__ == '__main__':
  main()



#add parent dir to find package. Only needed for source code build, pip install doesn't need it.
import os, inspect
currentdir = os.path.dirname(os.path.abspath(inspect.getfile(inspect.currentframe())))
parentdir = os.path.dirname(os.path.dirname(currentdir))
os.sys.path.insert(0, parentdir)

from pybullet_envs.bullet.racecarGymEnv import RacecarGymEnv
isDiscrete = False


def main():

  environment = RacecarGymEnv(renders=True, isDiscrete=isDiscrete)
  environment.reset()

  targetVelocitySlider = environment._p.addUserDebugParameter("wheelVelocity", -1, 1, 0)
  steeringSlider = environment._p.addUserDebugParameter("steering", -1, 1, 0)

  while (True):
    targetVelocity = environment._p.readUserDebugParameter(targetVelocitySlider)
    steeringAngle = environment._p.readUserDebugParameter(steeringSlider)
    if (isDiscrete):
      discreteAction = 0
      if (targetVelocity < -0.33):
        discreteAction = 0
      else:
        if (targetVelocity > 0.33):
          discreteAction = 6
        else:
          discreteAction = 3
      if (steeringAngle > -0.17):
        if (steeringAngle > 0.17):
          discreteAction = discreteAction + 2
        else:
          discreteAction = discreteAction + 1
      action = discreteAction
    else:
      action = [targetVelocity, steeringAngle]
    state, reward, done, info = environment.step(action)
    obs = environment.getExtendedObservation()
    print("obs")
    print(obs)


if __name__ == "__main__":
  main()



"""Generates a random terrain at Minitaur gym environment reset."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import os, inspect
currentdir = os.path.dirname(os.path.abspath(inspect.getfile(inspect.currentframe())))
parentdir = os.path.dirname(os.path.dirname(currentdir))
parentdir = os.path.dirname(os.path.dirname(parentdir))
os.sys.path.insert(0, parentdir)

import itertools
import math
import enum
import numpy as np

from pybullet_envs.minitaur.envs import env_randomizer_base

_GRID_LENGTH = 15
_GRID_WIDTH = 10
_MAX_SAMPLE_SIZE = 30
_MIN_BLOCK_DISTANCE = 0.7
_MAX_BLOCK_LENGTH = _MIN_BLOCK_DISTANCE
_MIN_BLOCK_LENGTH = _MAX_BLOCK_LENGTH / 2
_MAX_BLOCK_HEIGHT = 0.05
_MIN_BLOCK_HEIGHT = _MAX_BLOCK_HEIGHT / 2


class PoissonDisc2D(object):
  """Generates 2D points using Poisson disk sampling method.

  Implements the algorithm described in:
    http://www.cs.ubc.ca/~rbridson/docs/bridson-siggraph07-poissondisk.pdf
  Unlike the uniform sampling method that creates small clusters of points,
  Poisson disk method enforces the minimum distance between points and is more
  suitable for generating a spatial distribution of non-overlapping objects.
  """

  def __init__(self, grid_length, grid_width, min_radius, max_sample_size):
    """Initializes the algorithm.

    Args:
      grid_length: The length of the bounding square in which points are
        sampled.
      grid_width: The width of the bounding square in which points are
        sampled.
      min_radius: The minimum distance between any pair of points.
      max_sample_size: The maximum number of sample points around a active site.
        See details in the algorithm description.
    """
    self._cell_length = min_radius / math.sqrt(2)
    self._grid_length = grid_length
    self._grid_width = grid_width
    self._grid_size_x = int(grid_length / self._cell_length) + 1
    self._grid_size_y = int(grid_width / self._cell_length) + 1
    self._min_radius = min_radius
    self._max_sample_size = max_sample_size

    # Flattern the 2D grid as an 1D array. The grid is used for fast nearest
    # point searching.
    self._grid = [None] * self._grid_size_x * self._grid_size_y

    # Generate the first sample point and set it as an active site.
    first_sample = np.array(np.random.random_sample(2)) * [grid_length, grid_width]
    self._active_list = [first_sample]

    # Also store the sample point in the grid.
    self._grid[self._point_to_index_1d(first_sample)] = first_sample

  def _point_to_index_1d(self, point):
    """Computes the index of a point in the grid array.

    Args:
      point: A 2D point described by its coordinates (x, y).

    Returns:
      The index of the point within the self._grid array.
    """
    return self._index_2d_to_1d(self._point_to_index_2d(point))

  def _point_to_index_2d(self, point):
    """Computes the 2D index (aka cell ID) of a point in the grid.

    Args:
      point: A 2D point (list) described by its coordinates (x, y).

    Returns:
      x_index: The x index of the cell the point belongs to.
      y_index: The y index of the cell the point belongs to.
    """
    x_index = int(point[0] / self._cell_length)
    y_index = int(point[1] / self._cell_length)
    return x_index, y_index

  def _index_2d_to_1d(self, index2d):
    """Converts the 2D index to the 1D position in the grid array.

    Args:
      index2d: The 2D index of a point (aka the cell ID) in the grid.

    Returns:
      The 1D position of the cell within the self._grid array.
    """
    return index2d[0] + index2d[1] * self._grid_size_x

  def _is_in_grid(self, point):
    """Checks if the point is inside the grid boundary.

    Args:
      point: A 2D point (list) described by its coordinates (x, y).

    Returns:
      Whether the point is inside the grid.
    """
    return (0 <= point[0] < self._grid_length) and (0 <= point[1] < self._grid_width)

  def _is_in_range(self, index2d):
    """Checks if the cell ID is within the grid.

    Args:
      index2d: The 2D index of a point (aka the cell ID) in the grid.

    Returns:
      Whether the cell (2D index) is inside the grid.
    """

    return (0 <= index2d[0] < self._grid_size_x) and (0 <= index2d[1] < self._grid_size_y)

  def _is_close_to_existing_points(self, point):
    """Checks if the point is close to any already sampled (and stored) points.

    Args:
      point: A 2D point (list) described by its coordinates (x, y).

    Returns:
      True iff the distance of the point to any existing points is smaller than
      the min_radius
    """
    px, py = self._point_to_index_2d(point)
    # Now we can check nearby cells for existing points
    for neighbor_cell in itertools.product(xrange(px - 1, px + 2), xrange(py - 1, py + 2)):

      if not self._is_in_range(neighbor_cell):
        continue

      maybe_a_point = self._grid[self._index_2d_to_1d(neighbor_cell)]
      if maybe_a_point is not None and np.linalg.norm(maybe_a_point - point) < self._min_radius:
        return True

    return False

  def sample(self):
    """Samples new points around some existing point.

    Removes the sampling base point and also stores the new jksampled points if
    they are far enough from all existing points.
    """
    active_point = self._active_list.pop()
    for _ in xrange(self._max_sample_size):
      # Generate random points near the current active_point between the radius
      random_radius = np.random.uniform(self._min_radius, 2 * self._min_radius)
      random_angle = np.random.uniform(0, 2 * math.pi)

      # The sampled 2D points near the active point
      sample = random_radius * np.array([np.cos(random_angle),
                                         np.sin(random_angle)]) + active_point

      if not self._is_in_grid(sample):
        continue

      if self._is_close_to_existing_points(sample):
        continue

      self._active_list.append(sample)
      self._grid[self._point_to_index_1d(sample)] = sample

  def generate(self):
    """Generates the Poisson disc distribution of 2D points.

    Although the while loop looks scary, the algorithm is in fact O(N), where N
    is the number of cells within the grid. When we sample around a base point
    (in some base cell), new points will not be pushed into the base cell
    because of the minimum distance constraint. Once the current base point is
    removed, all future searches cannot start from within the same base cell.

    Returns:
      All sampled points. The points are inside the quare [0, grid_length] x [0,
      grid_width]
    """

    while self._active_list:
      self.sample()

    all_sites = []
    for p in self._grid:
      if p is not None:
        all_sites.append(p)

    return all_sites


class TerrainType(enum.Enum):
  """The randomzied terrain types we can use in the gym env."""
  RANDOM_BLOCKS = 1
  TRIANGLE_MESH = 2


class MinitaurTerrainRandomizer(env_randomizer_base.EnvRandomizerBase):
  """Generates an uneven terrain in the gym env."""

  def __init__(self,
               terrain_type=TerrainType.TRIANGLE_MESH,
               mesh_filename="robotics/reinforcement_learning/minitaur/envs/testdata/"
               "triangle_mesh_terrain/terrain9735.obj",
               mesh_scale=None):
    """Initializes the randomizer.

    Args:
      terrain_type: Whether to generate random blocks or load a triangle mesh.
      mesh_filename: The mesh file to be used. The mesh will only be loaded if
        terrain_type is set to TerrainType.TRIANGLE_MESH.
      mesh_scale: the scaling factor for the triangles in the mesh file.
    """
    self._terrain_type = terrain_type
    self._mesh_filename = mesh_filename
    self._mesh_scale = mesh_scale if mesh_scale else [1.0, 1.0, 0.3]

  def randomize_env(self, env):
    """Generate a random terrain for the current env.

    Args:
      env: A minitaur gym environment.
    """

    if self._terrain_type is TerrainType.TRIANGLE_MESH:
      self._load_triangle_mesh(env)
    if self._terrain_type is TerrainType.RANDOM_BLOCKS:
      self._generate_convex_blocks(env)

  def _load_triangle_mesh(self, env):
    """Represents the random terrain using a triangle mesh.

    It is possible for Minitaur leg to stuck at the common edge of two triangle
    pieces. To prevent this from happening, we recommend using hard contacts
    (or high stiffness values) for Minitaur foot in sim.

    Args:
      env: A minitaur gym environment.
    """
    env.pybullet_client.removeBody(env.ground_id)
    terrain_collision_shape_id = env.pybullet_client.createCollisionShape(
        shapeType=env.pybullet_client.GEOM_MESH,
        fileName=self._mesh_filename,
        flags=1,
        meshScale=self._mesh_scale)
    env.ground_id = env.pybullet_client.createMultiBody(
        baseMass=0, baseCollisionShapeIndex=terrain_collision_shape_id, basePosition=[0, 0, 0])

  def _generate_convex_blocks(self, env):
    """Adds random convex blocks to the flat ground.

    We use the Possion disk algorithm to add some random blocks on the ground.
    Possion disk algorithm sets the minimum distance between two sampling
    points, thus voiding the clustering effect in uniform N-D distribution.

    Args:
      env: A minitaur gym environment.

    """

    poisson_disc = PoissonDisc2D(_GRID_LENGTH, _GRID_WIDTH, _MIN_BLOCK_DISTANCE, _MAX_SAMPLE_SIZE)

    block_centers = poisson_disc.generate()

    for center in block_centers:
      # We want the blocks to be in front of the robot.
      shifted_center = np.array(center) - [2, _GRID_WIDTH / 2]

      # Do not place blocks near the point [0, 0], where the robot will start.
      if abs(shifted_center[0]) < 1.0 and abs(shifted_center[1]) < 1.0:
        continue
      half_length = np.random.uniform(_MIN_BLOCK_LENGTH, _MAX_BLOCK_LENGTH) / (2 * math.sqrt(2))
      half_height = np.random.uniform(_MIN_BLOCK_HEIGHT, _MAX_BLOCK_HEIGHT) / 2
      box_id = env.pybullet_client.createCollisionShape(
          env.pybullet_client.GEOM_BOX, halfExtents=[half_length, half_length, half_height])
      env.pybullet_client.createMultiBody(
          baseMass=0,
          baseCollisionShapeIndex=box_id,
          basePosition=[shifted_center[0], shifted_center[1], half_height])



"""Base class for simulation client."""

import enum
from typing import Text

GRAY = (0.3, 0.3, 0.3, 1)


class ClientType(enum.Enum):
  """Type of client."""
  BULLET = "pybullet"
  


class ClientMode(enum.Enum):
  """Client running mode."""

  # Client is being initialized, object is being loaded, teleported, etc.
  CONFIGURATION = 1

  # Client is in a mode that simulate motion according to physics and control.
  SIMULATION = 2


class WrongClientModeError(Exception):
  """Client mode does not meet expectation (e.g. load object in sim mode)."""


class BaseClient(object):
  """Base class for simulation client."""

  def __init__(self, client_type: Text = ""):
    self._client_type = client_type

    # Default to configuration mode.
    self._client_mode = ClientMode.CONFIGURATION

  @property
  def client_type(self) -> ClientType:
    return self._client_type

  def switch_mode(self, mode: ClientMode) -> bool:
    """Switches running mode of simulation client and return if mode changed."""
    if mode not in (ClientMode.CONFIGURATION, ClientMode.SIMULATION):
      raise ValueError(f"Invalid client mode {mode}.")
    if mode == self._client_mode:
      return False
    self._client_mode = mode
    return True

  @property
  def client_mode(self) -> ClientMode:
    """Returns current client mode."""
    return self._client_mode

  def _assert_in_configuration_mode(self, operation: Text = "this operation"):
    """Raises exception if client is not in configuration mode."""
    if self._client_mode != ClientMode.CONFIGURATION:
      raise WrongClientModeError(
          f"Sim client is expected to be in configuration mode for "
          f"{operation}.")





import sys, os
sys.path.append(os.path.dirname(__file__))
import pybullet

import gym


class Scene:
  "A base class for single- and multiplayer scenes"

  def __init__(self, bullet_client, gravity, timestep, frame_skip):
    self._p = bullet_client
    self.np_random, seed = gym.utils.seeding.np_random(None)
    self.timestep = timestep
    self.frame_skip = frame_skip

    self.dt = self.timestep * self.frame_skip
    self.cpp_world = World(self._p, gravity, timestep, frame_skip)

    self.test_window_still_open = True  # or never opened
    self.human_render_detected = False  # if user wants render("human"), we open test window

    self.multiplayer_robots = {}

  def test_window(self):
    "Call this function every frame, to see what's going on. Not necessary in learning."
    self.human_render_detected = True
    return self.test_window_still_open

  def actor_introduce(self, robot):
    "Usually after scene reset"
    if not self.multiplayer: return
    self.multiplayer_robots[robot.player_n] = robot

  def actor_is_active(self, robot):
    """
        Used by robots to see if they are free to exclusiveley put their HUD on the test window.
        Later can be used for click-focus robots.
        """
    return not self.multiplayer

  def episode_restart(self, bullet_client):
    "This function gets overridden by specific scene, to reset specific objects into their start positions"
    self.cpp_world.clean_everything()
    #self.cpp_world.test_window_history_reset()

  def global_step(self):
    """
        The idea is: apply motor torques for all robots, then call global_step(), then collect
        observations from robots using step() with the same action.
        """
    self.cpp_world.step(self.frame_skip)


class SingleRobotEmptyScene(Scene):
  multiplayer = False  # this class is used "as is" for InvertedPendulum, Reacher


class World:

  def __init__(self, bullet_client, gravity, timestep, frame_skip):
    self._p = bullet_client
    self.gravity = gravity
    self.timestep = timestep
    self.frame_skip = frame_skip
    self.numSolverIterations = 5
    self.clean_everything()

  def clean_everything(self):
    #p.resetSimulation()
    self._p.setGravity(0, 0, -self.gravity)
    self._p.setDefaultContactERP(0.9)
    #print("self.numSolverIterations=",self.numSolverIterations)
    self._p.setPhysicsEngineParameter(fixedTimeStep=self.timestep * self.frame_skip,
                                      numSolverIterations=self.numSolverIterations,
                                      numSubSteps=self.frame_skip)

  def step(self, frame_skip):
    self._p.stepSimulation()



# Lint as: python3
"""A scene containing only a planar floor."""

from typing import Sequence

import gin
from pybullet_envs.minitaur.envs_v2 import base_client
from pybullet_envs.minitaur.envs_v2.scenes import scene_base

_PLANE_URDF = (
    "plane.urdf")


@gin.configurable
class SimpleScene(scene_base.SceneBase):
  """A scene containing only a planar floor."""

  def build_scene(self, pybullet_client):
    super().build_scene(pybullet_client)

    visual_shape_id = self._pybullet_client.createVisualShape(
        shapeType=self._pybullet_client.GEOM_PLANE)
    collision_shape_id = self._pybullet_client.createCollisionShape(
        shapeType=self._pybullet_client.GEOM_PLANE)
    ground_id = self._pybullet_client.createMultiBody(
        baseMass=0,
        baseCollisionShapeIndex=collision_shape_id,
        baseVisualShapeIndex=visual_shape_id)
    self._pybullet_client.changeDynamics(ground_id, -1, lateralFriction=1.0)
    self.add_object(ground_id, scene_base.ObjectType.GROUND)

  @property
  def vectorized_map(self) -> Sequence[scene_base.Polygon]:
    """Returns vectorized map containing a list of polygon obstacles."""
    return []




from __future__ import absolute_import
from __future__ import division
from __future__ import print_function


def get_robot_base_position(robot):
  """Gets the base position of robot."""
  # TODO(b/151975607): Clean this after robot interface migration.
  if hasattr(robot, "GetBasePosition"):
    return robot.GetBasePosition()
  else:
    return robot.base_position


def get_robot_base_orientation(robot):
  """Gets the base orientation of robot."""
  # TODO(b/151975607): Clean this after robot interface migration.
  if hasattr(robot, "GetBaseOrientation"):
    return robot.GetBaseOrientation()
  else:
    return robot.base_orientation_quaternion


"""Utility functions for the Minitaur Raibert controller."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import math

UPPER_LEG_LEN = 0.112
LOWER_SHORT_LEG_LEN = 0.199
LOWER_LONG_LEG_LEN = 0.2315


def leg_pose_to_foot_position(leg_pose):
  """The forward kinematics."""
  l1 = UPPER_LEG_LEN
  l2 = LOWER_SHORT_LEG_LEN
  l3 = LOWER_LONG_LEG_LEN

  ext = leg_pose[1]
  alpha = math.asin(l1 * math.sin(ext) / l2)

  sw = -leg_pose[0]
  x = l3 * math.sin(alpha + sw) - l1 * math.sin(ext + sw)
  y = l3 * math.cos(alpha + sw) - l1 * math.cos(ext + sw)

  return (x, -y)


def foot_position_to_leg_pose(foot_position):
  """The inverse kinematics."""
  l1 = UPPER_LEG_LEN
  l2 = LOWER_SHORT_LEG_LEN
  l3 = LOWER_LONG_LEG_LEN

  x = foot_position[0]
  y = foot_position[1]

  assert y < 0
  hip_toe_sqr = x**2 + y**2
  cos_beta = (l1 * l1 + l3 * l3 - hip_toe_sqr) / (2 * l1 * l3)
  assert -1 <= cos_beta <= 1
  hip_ankle_sqr = l1 * l1 + l2 * l2 - 2 * l1 * l2 * cos_beta
  hip_ankle = math.sqrt(hip_ankle_sqr)
  cos_ext = -(l1 * l1 + hip_ankle_sqr - l2 * l2) / (2 * l1 * hip_ankle)
  ext = math.acos(cos_ext)

  hip_toe = math.sqrt(hip_toe_sqr)
  cos_theta = (hip_toe_sqr + hip_ankle_sqr -
               (l3 - l2)**2) / (2 * hip_ankle * hip_toe)

  assert cos_theta > 0
  theta = math.acos(cos_theta)
  sw = math.asin(x / hip_toe) - theta
  return (-sw, ext)


def extension_to_ankle_dist(extension):
  """Converts leg extension to ankle-hip distance in meters.

  The ankle is defined as the joint of the two lower links, which is different
  from the toe which is the tip of the longer lower limb.

  Args:
    extension: Float. the leg extension.

  Returns:
    Float. The hip to ankle distance in meters.

  """
  l1 = UPPER_LEG_LEN
  l2 = LOWER_SHORT_LEG_LEN
  alpha = math.asin(l1 / l2 * math.sin(extension))
  return l2 * math.cos(alpha) - l1 * math.cos(extension)


def ankle_dist_to_extension(dist):
  """Converts ankle-hip distance (meters) to extension."""
  l1 = UPPER_LEG_LEN
  l2 = LOWER_SHORT_LEG_LEN
  cos_extension = -(l1**2 + dist**2 - l2**2) / (2 * l1 * dist)
  return math.acos(cos_extension)



"""Contains APIs to create in sim camera images."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import logging
import enum
import gin
import numpy as np
from pybullet_envs.minitaur.vision import point_cloud_utils

_DEFAULT_TARGET_DISTANCE = 10
_DEFAULT_FOV = 60
_DEFAULT_OPENGL_FRUSTUM_FAR = 100.0
_DEFAULT_OPENGL_FRUSTUM_NEAR = 0.01
_PYBULLET_DEFAULT_PROJECTION_MATRIX = (1.0, 0.0, 0.0, 0.0, 0.0, 1.0, 0.0, 0.0,
                                       0.0, 0.0, -1.0000200271606445, -1.0, 0.0,
                                       0.0, -0.02000020071864128, 0.0)
# In the camera space, x axis points right, y axis points down and z axis points
# front. In the world space, x is front, y is left, and z is up. The following
# transformation rotate the camera coordinate to the world coordinate.
_CAMERA_Z_TO_WORLD_Z = (0.5, -0.5, 0.5, -0.5)


@gin.constants_from_enum
class CameraMode(enum.Enum):
  """Different modes that the camera can operate in."""

  RGB = 1
  DEPTH = 2
  RGBD = 3
  # The point cloud in the world frame, where z-axis is the up direction, x-axis
  # is the front direction, and y axis is the left direction.
  POINTCLOUD_WORLD_FRAME = 4
  # The point cloud in the robot's camera frame, where z-axis is the front
  # direction, x-axis is right direction, and y-axis is the downwards
  # direction.
  POINTCLOUD_ROBOT_FRAME = 5


def from_opengl_depth_to_distance(depth, near, far):
  return far * near / (far - (far - near) * depth)


def camera_intrinsics_to_projection_matrix(camera_intrinsics,
                                           width,
                                           height,
                                           far=_DEFAULT_OPENGL_FRUSTUM_FAR,
                                           near=_DEFAULT_OPENGL_FRUSTUM_NEAR):
  """Converts the camera intrinsics to OpenGL projection matrix.

  Args:
    camera_intrinsics: A CameraIntrinsics proto.
    width: Integer. The image width in pixels.
    height: Integer. The image height in pixels.
    far: Float. The far clipping plane distance.
    near: Float. The near clipping plane distance.

  Returns:
    An OpengGL projection matrix.

  """
  # first column
  projection_mat = [2 * camera_intrinsics.focal_length_px.x / width, 0, 0, 0]
  # second column
  projection_mat.extend(
      [0, 2 * camera_intrinsics.focal_length_px.y / height, 0, 0])
  # third column
  projection_mat.extend([
      2 * camera_intrinsics.principal_point_px.x / width - 1,
      2 * camera_intrinsics.principal_point_px.y / height - 1,
      -(far + near) / (far - near), -1
  ])
  # last column
  projection_mat.extend([0, 0, 2 * far * near / (near - far), 0])

  return tuple(projection_mat)


def create_camera_image(pybullet_client,
                        camera_position,
                        camera_orientation,
                        resolution,
                        projection_mat,
                        egl_render=False):
  """Creates a simulated camera image.

  Args:
    pybullet_client: A pybullet client.
    camera_position: A list of three floats. The absolute location of the camera
      in the simulated world.
    camera_orientation: A list of four floats. The orientation of the camera in
      world frame, in quaternion.
    resolution: A list of two integers. The horizontal and vertical resolution
      of the camera image in pixels.
    projection_mat: A list of 16 floats. The OpenGL projection matrix, in row
      major.
    egl_render: Whether to use EGL to render the images.

  Returns:
    A tuple containing the image resolution and the array for the sythesized RGB
    camera image.
  """

  orientation_mat = pybullet_client.getMatrixFromQuaternion(camera_orientation)

  # The first column in the orientation matrix.
  forward_vec = orientation_mat[::3]
  target_distance = _DEFAULT_TARGET_DISTANCE
  camera_target = [
      camera_position[0] + forward_vec[0] * target_distance,
      camera_position[1] + forward_vec[1] * target_distance,
      camera_position[2] + forward_vec[2] * target_distance
  ]

  # The third column in the orientation matrix. We assume camera up vector is
  # always [0, 0, 1] in its local frame.
  up_vec = orientation_mat[2::3]

  view_mat = pybullet_client.computeViewMatrix(camera_position, camera_target,
                                               up_vec)
  renderer = (
      pybullet_client.ER_BULLET_HARDWARE_OPENGL
      if egl_render else pybullet_client.ER_TINY_RENDERER)
  return pybullet_client.getCameraImage(
      resolution[0],
      resolution[1],
      viewMatrix=view_mat,
      projectionMatrix=projection_mat,
      renderer=renderer)

class MountedCamera(object):
  """A camera that is fixed to a robot's body part.

  Attributes:
    resolution: A 2-tuple (width, height) that represents the resolution of the
      camera.
    fov_degree: A floating point value that represents the field of view of the
      camera in the vertical direction. The unit is degree.
  """

  def __init__(self,
               pybullet_client,
               body_id,
               parent_link_id,
               relative_translation,
               relative_rotation,
               resolution,
               fov_degree=_DEFAULT_FOV,
               depth_lower_limit=_DEFAULT_OPENGL_FRUSTUM_NEAR,
               depth_upper_limit=_DEFAULT_OPENGL_FRUSTUM_FAR,
               stabilized=False,
               camera_intrinsics=None,
               camera_mode=CameraMode.RGB,
               normalize_rgb=False,
               egl_render=False):
    """Initializes the MounteeCamera class.

    Args:
      pybullet_client: A BulletClient instance.
      body_id: Integer. The unique body ID returned from loadURDF in pybullet.
      parent_link_id: Integer. The camera is rigidly attached to a body link
        specified by this ID. For example, camers may be mounted to the base or
        other moving parts of the robot.
      relative_translation: A list of three floats. The relative translation
        between the center of the camera and the link.
      relative_rotation: A list of four floats. The quaternion specifying the
        relative rotation of the camera.
      resolution: A list of two integers.
      fov_degree: The vertical field of view of this camera. Has no effect if
        camera_intrinsics is provided.
      depth_lower_limit: The lower limit of the depth camera.
      depth_upper_limit: The upper limit of the depth camera.
      stabilized: Weather the camera's roll and pitch should be stabilized. If
        turned on, only yaw rotation is applied to the camera.
      camera_intrinsics: A CameraIntrinsics proto instance.
      camera_mode: The mode of the camera. It can be set in the following
        choices [CameraMode.RGB, CameraMode.DEPTH, CameraMode.RGBD].
      normalize_rgb: Whether to normalize the output image pixel values to [-1,
        1].
      egl_render: Whether to use EGL to render the images when the X11 windows
        are not present.
    """
    self._pybullet_client = pybullet_client
    self._body_id = body_id
    self._parent_link_id = parent_link_id
    self._relative_translation = relative_translation
    self._relative_rotation = relative_rotation
    self._resolution = resolution
    self._stabilized = stabilized
    self._camera_mode = camera_mode
    self._fov_degree = fov_degree
    self._near = depth_lower_limit
    self._far = depth_upper_limit
    self._egl_render = egl_render
    self._normalize_rgb = normalize_rgb
    self._prev_depth = None
    width, height = resolution
    if camera_intrinsics is not None:
      self._projection_mat = camera_intrinsics_to_projection_matrix(
          camera_intrinsics, width, height, near=self._near, far=self._far)
    else:
      self._projection_mat = self._pybullet_client.computeProjectionMatrixFOV(
          fov_degree,
          float(width) / float(height), self._near, self._far)

  def __call__(self):
    return self.render_image()

  def transform_point_cloud_from_camera_to_world_frame(self, point_cloud):
    """Project a sample in the depth image to a 3D point in the world frame.

    Args:
      point_cloud: A numpy array of shape (height, width, 3) that represents
        a point cloud in the camera frame.

    Returns:
      The transformed point cloud in the world frame.
    """
    if self._parent_link_id == -1:
      camera_link_pos, camera_link_ori = (
          self._pybullet_client.getBasePositionAndOrientation(self._body_id))
    else:
      parent_link_state = self._pybullet_client.getLinkState(
          self._body_id, self._parent_link_id, computeForwardKinematics=True)
      camera_link_pos = parent_link_state[0]
      camera_link_ori = parent_link_state[1]

    camera_position_world, camera_orientation_world = (
        self._pybullet_client.multiplyTransforms(camera_link_pos,
                                                 camera_link_ori,
                                                 self._relative_translation,
                                                 self._relative_rotation))
    _, camera_space_to_world_orientation = (
        self._pybullet_client.multiplyTransforms([0, 0, 0],
                                                 camera_orientation_world,
                                                 [0, 0, 0],
                                                 _CAMERA_Z_TO_WORLD_Z))
    for i in range(point_cloud.shape[0]):
      for j in range(point_cloud.shape[1]):
        point_cloud[i, j, :] = (
            self._pybullet_client.multiplyTransforms(
                camera_position_world, camera_space_to_world_orientation,
                point_cloud[i, j, :], [0, 0, 0, 1])[0])
    return point_cloud

  def transform_point_cloud_from_camera_to_robot_frame(self, point_cloud):
    """Project a sample in the depth image to a 3D point in the robot frame.

    Args:
      point_cloud: A numpy array of shape (height, width, 3) that represents
        a point cloud in the camera frame.

    Returns:
      The transformed point cloud in the robot frame.
    """
    camera_space_to_robot_translation, camera_space_to_robot_orientation = (
        self._pybullet_client.multiplyTransforms(self._relative_translation,
                                                 self._relative_rotation,
                                                 [0, 0, 0],
                                                 _CAMERA_Z_TO_WORLD_Z))
    for i in range(point_cloud.shape[0]):
      for j in range(point_cloud.shape[1]):
        point_cloud[i, j, :] = (
            self._pybullet_client.multiplyTransforms(
                camera_space_to_robot_translation,
                camera_space_to_robot_orientation, point_cloud[i, j, :],
                [0, 0, 0, 1])[0])
    return point_cloud

  def render_image(self):
    """Retrieves the camera image.

    Returns:
      A simulated view from the camera's perspective.
    """
    pos = []
    ori = []
    if self._parent_link_id == -1:
      pos, ori = self._pybullet_client.getBasePositionAndOrientation(
          self._body_id)
    else:
      parent_link_state = self._pybullet_client.getLinkState(
          self._body_id, self._parent_link_id, computeForwardKinematics=True)
      pos = parent_link_state[0]
      ori = parent_link_state[1]
    if self._stabilized:
      [_, _, yaw] = self._pybullet_client.getEulerFromQuaternion(ori)
      ori = self._pybullet_client.getQuaternionFromEuler([0, 0, yaw])

    transform = self._pybullet_client.multiplyTransforms(
        pos, ori, self._relative_translation, self._relative_rotation)

    _, _, rgba, depth, _ = create_camera_image(
        pybullet_client=self._pybullet_client,
        camera_position=transform[0],
        camera_orientation=transform[1],
        resolution=self._resolution,
        projection_mat=self._projection_mat,
        egl_render=self._egl_render)
    if np.any(np.isnan(depth)):
      logging.info("depth contained nan: %s, replacing with prev_depth: %s",
                   depth, self._prev_depth)
      depth = self._prev_depth
    self._prev_depth = depth
    rgba = np.reshape(np.array(rgba, dtype=np.float32), (self._resolution[0], self._resolution[1], -1))

    # Converts from OpenGL depth map to a distance map (unit: meter).
    distances = from_opengl_depth_to_distance(
        np.reshape(np.array(depth), (self._resolution[0], self._resolution[1], -1)), self._near, self._far)
    if self._normalize_rgb:
      # Does not normalize the depth image.
      rgba = 2 * rgba / 255.0 - 1

    if self._camera_mode == CameraMode.RGB:
      return rgba[:, :, 0:3]  # Remove the alpha channel
    elif self._camera_mode == CameraMode.DEPTH:
      return distances[:, :, np.newaxis]
    elif self._camera_mode == CameraMode.RGBD:
      rgbd_image = rgba
      rgbd_image[:, :, 3] = distances
      return rgbd_image
    elif self._camera_mode == CameraMode.POINTCLOUD_WORLD_FRAME:
      point_cloud = point_cloud_utils.distance_map_to_point_cloud(
          distances, self._fov_degree / 180.0 * np.pi, distances.shape[1],
          distances.shape[0])
      point_cloud = self.transform_point_cloud_from_camera_to_world_frame(
          point_cloud)
      return point_cloud
    elif self._camera_mode == CameraMode.POINTCLOUD_ROBOT_FRAME:
      point_cloud = point_cloud_utils.distance_map_to_point_cloud(
          distances, self._fov_degree / 180.0 * np.pi, distances.shape[1],
          distances.shape[0])
      point_cloud = self.transform_point_cloud_from_camera_to_robot_frame(
          point_cloud)
      return point_cloud
    else:
      raise NotImplementedError("Camera mode {} is not implemented.".format(
          self._camera_mode))

  @property
  def resolution(self):
    return self._resolution

  @property
  def fov_degree(self):
    return self._fov_degree



import numpy as np
from mpi4py import MPI

ROOT_PROC_RANK = 0


def get_num_procs():
  return MPI.COMM_WORLD.Get_size()


def get_proc_rank():
  return MPI.COMM_WORLD.Get_rank()


def is_root_proc():
  rank = get_proc_rank()
  return rank == ROOT_PROC_RANK


def bcast(x):
  MPI.COMM_WORLD.Bcast(x, root=ROOT_PROC_RANK)
  return


def reduce_sum(x):
  return reduce_all(x, MPI.SUM)


def reduce_prod(x):
  return reduce_all(x, MPI.PROD)


def reduce_avg(x):
  buffer = reduce_sum(x)
  buffer /= get_num_procs()
  return buffer


def reduce_min(x):
  return reduce_all(x, MPI.MIN)


def reduce_max(x):
  return reduce_all(x, MPI.MAX)


def reduce_all(x, op):
  is_array = isinstance(x, np.ndarray)
  x_buf = x if is_array else np.array([x])
  buffer = np.zeros_like(x_buf)
  MPI.COMM_WORLD.Allreduce(x_buf, buffer, op=op)
  buffer = buffer if is_array else buffer[0]
  return buffer


def gather_all(x):
  is_array = isinstance(x, np.ndarray)
  x_buf = np.array([x])
  buffer = np.zeros_like(x_buf)
  buffer = np.repeat(buffer, get_num_procs(), axis=0)
  MPI.COMM_WORLD.Allgather(x_buf, buffer)
  buffer = list(buffer)
  return buffer



import numpy as np
try:
  import tensorflow.compat.v1 as tf
except Exception:
  import tensorflow as tf
from abc import abstractmethod

from pybullet_envs.deep_mimic.learning.rl_agent import RLAgent
from pybullet_utils.logger import Logger
from pybullet_envs.deep_mimic.learning.tf_normalizer import TFNormalizer


class TFAgent(RLAgent):
  RESOURCE_SCOPE = 'resource'
  SOLVER_SCOPE = 'solvers'

  def __init__(self, world, id, json_data):
    self.tf_scope = 'agent'
    self.graph = tf.Graph()
    self.sess = tf.Session(graph=self.graph)

    super().__init__(world, id, json_data)
    self._build_graph(json_data)
    self._init_normalizers()
    return

  def __del__(self):
    self.sess.close()
    return

  def save_model(self, out_path):
    with self.sess.as_default(), self.graph.as_default():
      try:
        save_path = self.saver.save(self.sess, out_path, write_meta_graph=False, write_state=False)
        Logger.print2('Model saved to: ' + save_path)
      except:
        Logger.print2("Failed to save model to: " + save_path)
    return

  def load_model(self, in_path):
    with self.sess.as_default(), self.graph.as_default():
      self.saver.restore(self.sess, in_path)
      self._load_normalizers()
      Logger.print2('Model loaded from: ' + in_path)
    return

  def _get_output_path(self):
    assert (self.output_dir != '')
    file_path = self.output_dir + '/agent' + str(self.id) + '_model.ckpt'
    return file_path

  def _get_int_output_path(self):
    assert (self.int_output_dir != '')
    file_path = self.int_output_dir + (
        '/agent{:d}_models/agent{:d}_int_model_{:010d}.ckpt').format(self.id, self.id, self.iter)
    return file_path

  def _build_graph(self, json_data):
    with self.sess.as_default(), self.graph.as_default():
      with tf.variable_scope(self.tf_scope):
        self._build_nets(json_data)

        with tf.variable_scope(self.SOLVER_SCOPE):
          self._build_losses(json_data)
          self._build_solvers(json_data)

        self._initialize_vars()
        self._build_saver()
    return

  def _init_normalizers(self):
    with self.sess.as_default(), self.graph.as_default():
      # update normalizers to sync the tensorflow tensors
      self.s_norm.update()
      self.g_norm.update()
      self.a_norm.update()
    return

  @abstractmethod
  def _build_nets(self, json_data):
    pass

  @abstractmethod
  def _build_losses(self, json_data):
    pass

  @abstractmethod
  def _build_solvers(self, json_data):
    pass

  def _tf_vars(self, scope=''):
    with self.sess.as_default(), self.graph.as_default():
      res = tf.get_collection(tf.GraphKeys.TRAINABLE_VARIABLES, scope=self.tf_scope + '/' + scope)
      assert len(res) > 0
    return res

  def _build_normalizers(self):
    with self.sess.as_default(), self.graph.as_default(), tf.variable_scope(self.tf_scope):
      with tf.variable_scope(self.RESOURCE_SCOPE):
        self.s_norm = TFNormalizer(self.sess, 's_norm', self.get_state_size(),
                                   self.world.env.build_state_norm_groups(self.id))
        state_offset = -self.world.env.build_state_offset(self.id)
        print("state_offset=", state_offset)
        state_scale = 1 / self.world.env.build_state_scale(self.id)
        print("state_scale=", state_scale)
        self.s_norm.set_mean_std(-self.world.env.build_state_offset(self.id),
                                 1 / self.world.env.build_state_scale(self.id))

        self.g_norm = TFNormalizer(self.sess, 'g_norm', self.get_goal_size(),
                                   self.world.env.build_goal_norm_groups(self.id))
        self.g_norm.set_mean_std(-self.world.env.build_goal_offset(self.id),
                                 1 / self.world.env.build_goal_scale(self.id))

        self.a_norm = TFNormalizer(self.sess, 'a_norm', self.get_action_size())
        self.a_norm.set_mean_std(-self.world.env.build_action_offset(self.id),
                                 1 / self.world.env.build_action_scale(self.id))
    return

  def _load_normalizers(self):
    self.s_norm.load()
    self.g_norm.load()
    self.a_norm.load()
    return

  def _update_normalizers(self):
    with self.sess.as_default(), self.graph.as_default():
      super()._update_normalizers()
    return

  def _initialize_vars(self):
    self.sess.run(tf.global_variables_initializer())
    return

  def _build_saver(self):
    vars = self._get_saver_vars()
    self.saver = tf.train.Saver(vars, max_to_keep=0)
    return

  def _get_saver_vars(self):
    with self.sess.as_default(), self.graph.as_default():
      vars = tf.get_collection(tf.GraphKeys.GLOBAL_VARIABLES, scope=self.tf_scope)
      vars = [v for v in vars if '/' + self.SOLVER_SCOPE + '/' not in v.name]
      #vars = [v for v in vars if '/target/' not in v.name]
      assert len(vars) > 0
    return vars

  def _weight_decay_loss(self, scope):
    vars = self._tf_vars(scope)
    vars_no_bias = [v for v in vars if 'bias' not in v.name]
    loss = tf.add_n([tf.nn.l2_loss(v) for v in vars_no_bias])
    return loss

  def _train(self):
    with self.sess.as_default(), self.graph.as_default():
      super()._train()
    return



from . import enjoy_TF_AntBulletEnv_v0_2017may
from . import enjoy_TF_HalfCheetahBulletEnv_v0_2017may
from . import enjoy_TF_AntBulletEnv_v0_2017may
from . import enjoy_TF_HopperBulletEnv_v0_2017may
from . import enjoy_TF_HumanoidBulletEnv_v0_2017may
from . import enjoy_TF_InvertedDoublePendulumBulletEnv_v0_2017may
from . import enjoy_TF_InvertedPendulumBulletEnv_v0_2017may
from . import enjoy_TF_InvertedPendulumSwingupBulletEnv_v0_2017may
from . import enjoy_TF_Walker2DBulletEnv_v0_2017may
from . import kukaGymEnvTest
from . import racecarGymEnvTest
from . import racecarZEDGymEnvTest



import pybullet as p
import pybullet_data

p.connect(p.GUI)
p.setAdditionalSearchPath(pybullet_data.getDataPath())
plane = p.loadURDF("plane.urdf")
visualData = p.getVisualShapeData(plane, p.VISUAL_SHAPE_DATA_TEXTURE_UNIQUE_IDS)
print(visualData)
curTexUid = visualData[0][8]
print(curTexUid)
texUid = p.loadTexture("tex256.png")
print("texUid=", texUid)

p.changeVisualShape(plane, -1, textureUniqueId=texUid)

for i in range(100):
  p.getCameraImage(320, 200)
p.changeVisualShape(plane, -1, textureUniqueId=curTexUid)

for i in range(100):
  p.getCameraImage(320, 200)



from pybullet_utils import bullet_client
import time
import math
import motion_capture_data
from pybullet_envs.deep_mimic.env import humanoid_stable_pd
import pybullet_data
import pybullet as p1
import humanoid_pose_interpolator
import numpy as np

pybullet_client = bullet_client.BulletClient(connection_mode=p1.GUI)

pybullet_client.setAdditionalSearchPath(pybullet_data.getDataPath())
z2y = pybullet_client.getQuaternionFromEuler([-math.pi * 0.5, 0, 0])
#planeId = pybullet_client.loadURDF("plane.urdf",[0,0,0],z2y)
planeId = pybullet_client.loadURDF("plane_implicit.urdf", [0, 0, 0],
                                   z2y,
                                   useMaximalCoordinates=True)
pybullet_client.changeDynamics(planeId, linkIndex=-1, lateralFriction=0.9)
#print("planeId=",planeId)

pybullet_client.configureDebugVisualizer(pybullet_client.COV_ENABLE_Y_AXIS_UP, 1)
pybullet_client.setGravity(0, -9.8, 0)

pybullet_client.setPhysicsEngineParameter(numSolverIterations=10)

mocapData = motion_capture_data.MotionCaptureData()
#motionPath = pybullet_data.getDataPath()+"/data/motions/humanoid3d_walk.txt"
motionPath = pybullet_data.getDataPath() + "/data/motions/humanoid3d_backflip.txt"
mocapData.Load(motionPath)
timeStep = 1. / 600
useFixedBase = False
humanoid = humanoid_stable_pd.HumanoidStablePD(pybullet_client, mocapData, timeStep, useFixedBase)
isInitialized = True

pybullet_client.setTimeStep(timeStep)
pybullet_client.setPhysicsEngineParameter(numSubSteps=2)
timeId = pybullet_client.addUserDebugParameter("time", 0, 10, 0)


def isKeyTriggered(keys, key):
  o = ord(key)
  if o in keys:
    return keys[ord(key)] & pybullet_client.KEY_WAS_TRIGGERED
  return False


animating = False
singleStep = False
humanoid.resetPose()
t = 0
while (1):

  keys = pybullet_client.getKeyboardEvents()
  #print(keys)
  if isKeyTriggered(keys, ' '):
    animating = not animating

  if isKeyTriggered(keys, 'b'):
    singleStep = True

  if animating or singleStep:

    singleStep = False
    #t = pybullet_client.readUserDebugParameter(timeId)
    #print("t=",t)
    for i in range(1):

      #print("t=", t)
      humanoid.setSimTime(t)

      humanoid.computePose(humanoid._frameFraction)
      pose = humanoid._poseInterpolator
      #humanoid.initializePose(pose=pose, phys_model = humanoid._sim_model, initBase=True, initializeVelocity=True)
      #humanoid.resetPose()

      desiredPose = humanoid.computePose(humanoid._frameFraction)
      
      #desiredPose = desiredPose.GetPose()
      #curPose = HumanoidPoseInterpolator()
      #curPose.reset()
      
      s = humanoid.getState()
      #np.savetxt("pb_record_state_s.csv", s, delimiter=",")
      maxForces = [
          0, 0, 0, 0, 0, 0, 0, 200, 200, 200, 200, 50, 50, 50, 50, 200, 200, 200, 200, 150, 90, 90,
          90, 90, 100, 100, 100, 100, 60, 200, 200, 200, 200, 150, 90, 90, 90, 90, 100, 100, 100,
          100, 60
      ]
      
      usePythonStablePD = False
      if usePythonStablePD:
        taus = humanoid.computePDForces(desiredPose, desiredVelocities=None, maxForces=maxForces)
        #Print("taus=",taus)
        humanoid.applyPDForces(taus)
      else:
        humanoid.computeAndApplyPDForces(desiredPose,maxForces=maxForces)

      pybullet_client.stepSimulation()
      t += 1. / 600.

  time.sleep(1. / 600.)



"""State estimator."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from typing import Any, Sequence

import gin
import numpy as np

from pybullet_envs.minitaur.agents.baseline_controller import state_estimator
from pybullet_envs.minitaur.robots.safety.python import moving_window_filter

_DEFAULT_WINDOW_SIZE = 20


@gin.configurable
class COMVelocityEstimator(state_estimator.StateEstimatorBase):
  """Estimate the CoM velocity using on board sensors.


  Requires knowledge about the base velocity in world frame, which for example
  can be obtained from a MoCap system. This estimator will filter out the high
  frequency noises in the velocity so the results can be used with controllers
  reliably.

  """

  def __init__(
      self,
      robot: Any,
      window_size: int = _DEFAULT_WINDOW_SIZE,
  ):
    self._robot = robot
    self._window_size = window_size
    self.reset(0)

  @property
  def com_velocity_body_yaw_aligned_frame(self) -> Sequence[float]:
    """The base velocity projected in the body aligned inertial frame.

    The body aligned frame is a intertia frame that coincides with the body
    frame, but has a zero relative velocity/angular velocity to the world frame.

    Returns:
      The com velocity in body aligned frame.
    """
    return self._com_velocity_body_yaw_aligned_frame

  @property
  def com_velocity_world_frame(self) -> Sequence[float]:
    return self._com_velocity_world_frame

  def reset(self, current_time):
    del current_time
    # We use a moving window filter to reduce the noise in velocity estimation.
    self._velocity_filter_x = moving_window_filter.MovingWindowFilter(
        window_size=self._window_size)
    self._velocity_filter_y = moving_window_filter.MovingWindowFilter(
        window_size=self._window_size)
    self._velocity_filter_z = moving_window_filter.MovingWindowFilter(
        window_size=self._window_size)
    self._com_velocity_world_frame = np.array((0, 0, 0))
    self._com_velocity_body_yaw_aligned_frame = np.array((0, 0, 0))

  def update(self, current_time):
    del current_time
    velocity = self._robot.base_velocity

    vx = self._velocity_filter_x.CalculateAverage(velocity[0])
    vy = self._velocity_filter_y.CalculateAverage(velocity[1])
    vz = self._velocity_filter_z.CalculateAverage(velocity[2])
    self._com_velocity_world_frame = np.array((vx, vy, vz))

    base_orientation = self._robot.base_orientation_quaternion
    _, inverse_rotation = self._robot.pybullet_client.invertTransform(
        (0, 0, 0), base_orientation)

    self._com_velocity_body_yaw_aligned_frame, _ = (
        self._robot.pybullet_client.multiplyTransforms(
            (0, 0, 0), inverse_rotation, self._com_velocity_world_frame,
            (0, 0, 0, 1)))



#add parent dir to find package. Only needed for source code build, pip install doesn't need it.
import os, inspect
currentdir = os.path.dirname(os.path.abspath(inspect.getfile(inspect.currentframe())))
parentdir = os.path.dirname(os.path.dirname(currentdir))
os.sys.path.insert(0, parentdir)

from pybullet_envs.bullet.kukaGymEnv import KukaGymEnv
import time


def main():

  environment = KukaGymEnv(renders=True, isDiscrete=False, maxSteps=10000000)

  motorsIds = []
  #motorsIds.append(environment._p.addUserDebugParameter("posX",0.4,0.75,0.537))
  #motorsIds.append(environment._p.addUserDebugParameter("posY",-.22,.3,0.0))
  #motorsIds.append(environment._p.addUserDebugParameter("posZ",0.1,1,0.2))
  #motorsIds.append(environment._p.addUserDebugParameter("yaw",-3.14,3.14,0))
  #motorsIds.append(environment._p.addUserDebugParameter("fingerAngle",0,0.3,.3))

  dv = 0.01
  motorsIds.append(environment._p.addUserDebugParameter("posX", -dv, dv, 0))
  motorsIds.append(environment._p.addUserDebugParameter("posY", -dv, dv, 0))
  motorsIds.append(environment._p.addUserDebugParameter("posZ", -dv, dv, 0))
  motorsIds.append(environment._p.addUserDebugParameter("yaw", -dv, dv, 0))
  motorsIds.append(environment._p.addUserDebugParameter("fingerAngle", 0, 0.3, .3))

  done = False
  while (not done):

    action = []
    for motorId in motorsIds:
      action.append(environment._p.readUserDebugParameter(motorId))

    state, reward, done, info = environment.step2(action)
    obs = environment.getExtendedObservation()


if __name__ == "__main__":
  main()



"""Utility functions to calculate Mini Cheetah's pose and motor angles."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import attr

MINI_CHEETAH_DEFAULT_ABDUCTION_ANGLE = 0
MINI_CHEETAH_DEFAULT_HIP_ANGLE = -1.2
MINI_CHEETAH_DEFAULT_KNEE_ANGLE = 2.1


@attr.s
class MiniCheetahPose(object):
  """Default pose of the Laikago.

    Leg order:
    0 -> Front Right.
    1 -> Front Left.
    2 -> Rear Right.
    3 -> Rear Left.
  """
  abduction_angle_0 = attr.ib(type=float, default=0)
  hip_angle_0 = attr.ib(type=float, default=0)
  knee_angle_0 = attr.ib(type=float, default=0)
  abduction_angle_1 = attr.ib(type=float, default=0)
  hip_angle_1 = attr.ib(type=float, default=0)
  knee_angle_1 = attr.ib(type=float, default=0)
  abduction_angle_2 = attr.ib(type=float, default=0)
  hip_angle_2 = attr.ib(type=float, default=0)
  knee_angle_2 = attr.ib(type=float, default=0)
  abduction_angle_3 = attr.ib(type=float, default=0)
  hip_angle_3 = attr.ib(type=float, default=0)
  knee_angle_3 = attr.ib(type=float, default=0)



"""This file implements an accurate motor model."""

import numpy as np

from pybullet_envs.minitaur.robots import robot_config

VOLTAGE_CLIPPING = 50
# TODO(b/73728631): Clamp the pwm signal instead of the OBSERVED_TORQUE_LIMIT.
OBSERVED_TORQUE_LIMIT = 5.7
MOTOR_VOLTAGE = 16.0
MOTOR_RESISTANCE = 0.186
MOTOR_TORQUE_CONSTANT = 0.0954
MOTOR_VISCOUS_DAMPING = 0
MOTOR_SPEED_LIMIT = MOTOR_VOLTAGE / (
    MOTOR_VISCOUS_DAMPING + MOTOR_TORQUE_CONSTANT)
NUM_MOTORS = 8
MOTOR_POS_LB = 0.5
MOTOR_POS_UB = 2.5


class MotorModel(object):
  """The accurate motor model, which is based on the physics of DC motors.

  The motor model support two types of control: position control and torque
  control. In position control mode, a desired motor angle is specified, and a
  torque is computed based on the internal motor model. When the torque control
  is specified, a pwm signal in the range of [-1.0, 1.0] is converted to the
  torque.

  The internal motor model takes the following factors into consideration:
  pd gains, viscous friction, back-EMF voltage and current-torque profile.
  """

  def __init__(self,
               kp=1.2,
               kd=0,
               torque_limits=None,
               motor_control_mode=robot_config.MotorControlMode.POSITION):
    self._kp = kp
    self._kd = kd
    self._torque_limits = torque_limits
    self._motor_control_mode = motor_control_mode
    self._resistance = MOTOR_RESISTANCE
    self._voltage = MOTOR_VOLTAGE
    self._torque_constant = MOTOR_TORQUE_CONSTANT
    self._viscous_damping = MOTOR_VISCOUS_DAMPING
    self._current_table = [0, 10, 20, 30, 40, 50, 60]
    self._torque_table = [0, 1, 1.9, 2.45, 3.0, 3.25, 3.5]
    self._strength_ratios = [1.0] * NUM_MOTORS

  def set_strength_ratios(self, ratios):
    """Set the strength of each motors relative to the default value.

    Args:
      ratios: The relative strength of motor output. A numpy array ranging from
        0.0 to 1.0.
    """
    self._strength_ratios = np.array(ratios)

  def set_motor_gains(self, kp, kd):
    """Set the gains of all motors.

    These gains are PD gains for motor positional control. kp is the
    proportional gain and kd is the derivative gain.

    Args:
      kp: proportional gain of the motors.
      kd: derivative gain of the motors.
    """
    self._kp = kp
    self._kd = kd

  def set_voltage(self, voltage):
    self._voltage = voltage

  def get_voltage(self):
    return self._voltage

  def set_viscous_damping(self, viscous_damping):
    self._viscous_damping = viscous_damping

  def get_viscous_dampling(self):
    return self._viscous_damping

  def convert_to_torque(self,
                        motor_commands,
                        motor_angle,
                        motor_velocity,
                        true_motor_velocity,
                        motor_control_mode=None):
    """Convert the commands (position control or pwm control) to torque.

    Args:
      motor_commands: The desired motor angle if the motor is in position
        control mode. The pwm signal if the motor is in torque control mode.
      motor_angle: The motor angle observed at the current time step. It is
        actually the true motor angle observed a few milliseconds ago (pd
        latency).
      motor_velocity: The motor velocity observed at the current time step, it
        is actually the true motor velocity a few milliseconds ago (pd latency).
      true_motor_velocity: The true motor velocity. The true velocity is used to
        compute back EMF voltage and viscous damping.
      motor_control_mode: A MotorControlMode enum.

    Returns:
      actual_torque: The torque that needs to be applied to the motor.
      observed_torque: The torque observed by the sensor.
    """
    if not motor_control_mode:
      motor_control_mode = self._motor_control_mode

    if (motor_control_mode is robot_config.MotorControlMode.TORQUE) or (
        motor_control_mode is robot_config.MotorControlMode.HYBRID):
      raise ValueError(
          "{} is not a supported motor control mode".format(motor_control_mode))

    kp = self._kp
    kd = self._kd

    if motor_control_mode is robot_config.MotorControlMode.PWM:
      # The following implements a safety controller that softly enforces the
      # joint angles to remain within safe region: If PD controller targeting
      # the positive (negative) joint limit outputs a negative (positive)
      # signal, the corresponding joint violates the joint constraint, so
      # we should add the PD output to motor_command to bring it back to the
      # safe region.
      pd_max = -1 * kp * (motor_angle - MOTOR_POS_UB) - kd / 2. * motor_velocity
      pd_min = -1 * kp * (motor_angle - MOTOR_POS_LB) - kd / 2. * motor_velocity
      pwm = motor_commands + np.minimum(pd_max, 0) + np.maximum(pd_min, 0)
    else:
      pwm = -1 * kp * (motor_angle - motor_commands) - kd * motor_velocity
    pwm = np.clip(pwm, -1.0, 1.0)
    return self._convert_to_torque_from_pwm(pwm, true_motor_velocity)

  def _convert_to_torque_from_pwm(self, pwm, true_motor_velocity):
    """Convert the pwm signal to torque.

    Args:
      pwm: The pulse width modulation.
      true_motor_velocity: The true motor velocity at the current moment. It is
        used to compute the back EMF voltage and the viscous damping.

    Returns:
      actual_torque: The torque that needs to be applied to the motor.
      observed_torque: The torque observed by the sensor.
    """
    observed_torque = np.clip(
        self._torque_constant *
        (np.asarray(pwm) * self._voltage / self._resistance),
        -OBSERVED_TORQUE_LIMIT, OBSERVED_TORQUE_LIMIT)
    if self._torque_limits is not None:
      observed_torque = np.clip(observed_torque, -1.0 * self._torque_limits,
                                self._torque_limits)

    # Net voltage is clipped at 50V by diodes on the motor controller.
    voltage_net = np.clip(
        np.asarray(pwm) * self._voltage -
        (self._torque_constant + self._viscous_damping) *
        np.asarray(true_motor_velocity), -VOLTAGE_CLIPPING, VOLTAGE_CLIPPING)
    current = voltage_net / self._resistance
    current_sign = np.sign(current)
    current_magnitude = np.absolute(current)
    # Saturate torque based on empirical current relation.
    actual_torque = np.interp(current_magnitude, self._current_table,
                              self._torque_table)
    actual_torque = np.multiply(current_sign, actual_torque)
    actual_torque = np.multiply(self._strength_ratios, actual_torque)
    if self._torque_limits is not None:
      actual_torque = np.clip(actual_torque, -1.0 * self._torque_limits,
                              self._torque_limits)
    return actual_torque, observed_torque



import pybullet as p
p.connect(p.GUI)
p.loadURDF("combined.urdf", useFixedBase=True)

#for j in range (p.getNumJoints(0)):
#	p.setJointMotorControl2(0,j,p.VELOCITY_CONTROL,targetVelocity=0.1)
p.setRealTimeSimulation(1)
while (p.isConnected()):
  p.getCameraImage(320, 200)
  import time
  time.sleep(1. / 240.)



# Lint as: python3
"""Defines the LaikagoKinematic robot related constants and URDF specs."""

import collections
import gin

LAIKAGO_KINEMATIC_URDF_PATH = "robotics/reinforcement_learning/minitaur/robots/data/urdf/laikago/laikago_toes_zup_kinematic.urdf"
INIT_POSITION = (0, 0, 0.5)
INIT_ORIENTATION_QUAT = (0, 0, 0, 1)
INIT_ORIENTATION_RPY = (0, 0, 0)

NUM_LEGS = 4
# TODO(b/153405332): Use link name instead of joint name to identify the
# base link.
BASE_NAMES = ("kinematic_drive_joint_th",)

JOINT_NAMES = (
    "kinematic_drive_joint_x",
    "kinematic_drive_joint_y",
    "kinematic_drive_joint_th",
    # front right leg
    "FR_hip_motor_2_chassis_joint",
    "FR_upper_leg_2_hip_motor_joint",
    "FR_lower_leg_2_upper_leg_joint",
    # front left leg
    "FL_hip_motor_2_chassis_joint",
    "FL_upper_leg_2_hip_motor_joint",
    "FL_lower_leg_2_upper_leg_joint",
    # rear right leg
    "RR_hip_motor_2_chassis_joint",
    "RR_upper_leg_2_hip_motor_joint",
    "RR_lower_leg_2_upper_leg_joint",
    # rear left leg
    "RL_hip_motor_2_chassis_joint",
    "RL_upper_leg_2_hip_motor_joint",
    "RL_lower_leg_2_upper_leg_joint",
)

# A default joint pose where the arm is tucked near the base, and head looking
# forward.
INIT_ABDUCTION_ANGLE = 0
INIT_HIP_ANGLE = 0.67
INIT_KNEE_ANGLE = -1.25

# Note this matches the Laikago SDK/control convention, but is different from
# URDF's internal joint angles which needs to be computed using the joint
# offsets and directions. The conversion formula is (sdk_joint_angle + offset) *
# joint direction.
INIT_JOINT_ANGLES = collections.OrderedDict(
    zip(JOINT_NAMES, (0, 0, 0) +
        (INIT_ABDUCTION_ANGLE, INIT_HIP_ANGLE, INIT_KNEE_ANGLE) * NUM_LEGS))

# Used to convert the robot SDK joint angles to URDF joint angles.
JOINT_DIRECTIONS = collections.OrderedDict(
    zip(JOINT_NAMES, (1, 1, 1, -1, 1, 1, 1, 1, 1, -1, 1, 1, 1, 1, 1)))

HIP_JOINT_OFFSET = 0.0
UPPER_LEG_JOINT_OFFSET = -0.6
KNEE_JOINT_OFFSET = 0.66

# Used to convert the robot SDK joint angles to URDF joint angles.
JOINT_OFFSETS = collections.OrderedDict(
    zip(JOINT_NAMES, [0, 0, 0] +
        [HIP_JOINT_OFFSET, UPPER_LEG_JOINT_OFFSET, KNEE_JOINT_OFFSET] *
        NUM_LEGS))

LEG_NAMES = (
    "front_right",
    "front_left",
    "rear_right",
    "rear_left",
)

LEG_ORDER = (
    "front_right",
    "front_left",
    "back_right",
    "back_left",
)

END_EFFECTOR_NAMES = (
    "jtoeFR",
    "jtoeFL",
    "jtoeRR",
    "jtoeRL",
)

MOTOR_NAMES = JOINT_NAMES
MOTOR_GROUP = collections.OrderedDict((("body_motors", JOINT_NAMES[3:]),))

# Add the gin constants to be used for gin binding in config.
gin.constant("laikago_kinematic_constants.LAIKAGO_KINEMATIC_URDF_PATH",
             LAIKAGO_KINEMATIC_URDF_PATH)
gin.constant("laikago_kinematic_constants.INIT_POSITION", INIT_POSITION)
gin.constant("laikago_kinematic_constants.INIT_ORIENTATION_QUAT",
             INIT_ORIENTATION_QUAT)
gin.constant("laikago_kinematic_constants.INIT_ORIENTATION_RPY",
             INIT_ORIENTATION_RPY)
gin.constant("laikago_kinematic_constants.BASE_NAMES", BASE_NAMES)
gin.constant("laikago_kinematic_constants.INIT_JOINT_ANGLES", INIT_JOINT_ANGLES)
gin.constant("laikago_kinematic_constants.JOINT_DIRECTIONS", JOINT_DIRECTIONS)
gin.constant("laikago_kinematic_constants.JOINT_OFFSETS", JOINT_OFFSETS)
gin.constant("laikago_kinematic_constants.MOTOR_NAMES", MOTOR_NAMES)
gin.constant("laikago_kinematic_constants.END_EFFECTOR_NAMES",
             END_EFFECTOR_NAMES)
gin.constant("laikago_kinematic_constants.MOTOR_GROUP", MOTOR_GROUP)



from .scene_stadium import SinglePlayerStadiumScene
from .env_bases import MJCFBaseBulletEnv
import numpy as np
import pybullet
from robot_locomotors import Hopper, Walker2D, HalfCheetah, Ant, Humanoid, HumanoidFlagrun, HumanoidFlagrunHarder


class WalkerBaseBulletEnv(MJCFBaseBulletEnv):

  def __init__(self, robot, render=False):
    # print("WalkerBase::__init__ start")
    self.camera_x = 0
    self.walk_target_x = 1e3  # kilometer away
    self.walk_target_y = 0
    self.stateId = -1
    MJCFBaseBulletEnv.__init__(self, robot, render)


  def create_single_player_scene(self, bullet_client):
    self.stadium_scene = SinglePlayerStadiumScene(bullet_client,
                                                  gravity=9.8,
                                                  timestep=0.0165 / 4,
                                                  frame_skip=4)
    return self.stadium_scene


  def reset(self):
    if (self.stateId >= 0):
      #print("restoreState self.stateId:",self.stateId)
      self._p.restoreState(self.stateId)

    r = MJCFBaseBulletEnv.reset(self)
    self._p.configureDebugVisualizer(pybullet.COV_ENABLE_RENDERING, 0)

    self.parts, self.jdict, self.ordered_joints, self.robot_body = self.robot.addToScene(
        self._p, self.stadium_scene.ground_plane_mjcf)
    self.ground_ids = set([(self.parts[f].bodies[self.parts[f].bodyIndex],
                            self.parts[f].bodyPartIndex) for f in self.foot_ground_object_names])
    self._p.configureDebugVisualizer(pybullet.COV_ENABLE_RENDERING, 1)
    if (self.stateId < 0):
      self.stateId = self._p.saveState()
      #print("saving state self.stateId:",self.stateId)

    return r

  def _isDone(self):
    return self._alive < 0

  def move_robot(self, init_x, init_y, init_z):
    "Used by multiplayer stadium to move sideways, to another running lane."
    self.cpp_robot.query_position()
    pose = self.cpp_robot.root_part.pose()
    pose.move_xyz(
        init_x, init_y, init_z
    )  # Works because robot loads around (0,0,0), and some robots have z != 0 that is left intact
    self.cpp_robot.set_pose(pose)

  electricity_cost = -2.0  # cost for using motors -- this parameter should be carefully tuned against reward for making progress, other values less improtant
  stall_torque_cost = -0.1  # cost for running electric current through a motor even at zero rotational speed, small
  foot_collision_cost = -1.0  # touches another leg, or other objects, that cost makes robot avoid smashing feet into itself
  foot_ground_object_names = set(["floor"])  # to distinguish ground and other objects
  joints_at_limit_cost = -0.1  # discourage stuck joints

  def step(self, a):
    if not self.scene.multiplayer:  # if multiplayer, action first applied to all robots, then global step() called, then _step() for all robots with the same actions
      self.robot.apply_action(a)
      self.scene.global_step()

    state = self.robot.calc_state()  # also calculates self.joints_at_limit

    self._alive = float(
        self.robot.alive_bonus(
            state[0] + self.robot.initial_z,
            self.robot.body_rpy[1]))  # state[0] is body height above ground, body_rpy[1] is pitch
    done = self._isDone()
    if not np.isfinite(state).all():
      print("~INF~", state)
      done = True

    potential_old = self.potential
    self.potential = self.robot.calc_potential()
    progress = float(self.potential - potential_old)

    feet_collision_cost = 0.0
    for i, f in enumerate(
        self.robot.feet
    ):  # TODO: Maybe calculating feet contacts could be done within the robot code
      contact_ids = set((x[2], x[4]) for x in f.contact_list())
      #print("CONTACT OF '%d' WITH %d" % (contact_ids, ",".join(contact_names)) )
      if (self.ground_ids & contact_ids):
        #see Issue 63: https://github.com/openai/roboschool/issues/63
        #feet_collision_cost += self.foot_collision_cost
        self.robot.feet_contact[i] = 1.0
      else:
        self.robot.feet_contact[i] = 0.0

    electricity_cost = self.electricity_cost * float(np.abs(a * self.robot.joint_speeds).mean(
    ))  # let's assume we have DC motor with controller, and reverse current braking
    electricity_cost += self.stall_torque_cost * float(np.square(a).mean())

    joints_at_limit_cost = float(self.joints_at_limit_cost * self.robot.joints_at_limit)
    debugmode = 0
    if (debugmode):
      print("alive=")
      print(self._alive)
      print("progress")
      print(progress)
      print("electricity_cost")
      print(electricity_cost)
      print("joints_at_limit_cost")
      print(joints_at_limit_cost)
      print("feet_collision_cost")
      print(feet_collision_cost)

    self.rewards = [
        self._alive, progress, electricity_cost, joints_at_limit_cost, feet_collision_cost
    ]
    if (debugmode):
      print("rewards=")
      print(self.rewards)
      print("sum rewards")
      print(sum(self.rewards))
    self.HUD(state, a, done)
    self.reward += sum(self.rewards)

    return state, sum(self.rewards), bool(done), {}

  def camera_adjust(self):
    x, y, z = self.robot.body_real_xyz

    self.camera_x = x
    self.camera.move_and_look_at(self.camera_x, y , 1.4, x, y, 1.0)


class HopperBulletEnv(WalkerBaseBulletEnv):

  def __init__(self, render=False):
    self.robot = Hopper()
    WalkerBaseBulletEnv.__init__(self, self.robot, render)


class Walker2DBulletEnv(WalkerBaseBulletEnv):

  def __init__(self, render=False):
    self.robot = Walker2D()
    WalkerBaseBulletEnv.__init__(self, self.robot, render)


class HalfCheetahBulletEnv(WalkerBaseBulletEnv):

  def __init__(self, render=False):
    self.robot = HalfCheetah()
    WalkerBaseBulletEnv.__init__(self, self.robot, render)

  def _isDone(self):
    return False


class AntBulletEnv(WalkerBaseBulletEnv):

  def __init__(self, render=False):
    self.robot = Ant()
    WalkerBaseBulletEnv.__init__(self, self.robot, render)


class HumanoidBulletEnv(WalkerBaseBulletEnv):

  def __init__(self, robot=None, render=False):
    if robot is None:
      self.robot = Humanoid()
    else:
      self.robot = robot
    WalkerBaseBulletEnv.__init__(self, self.robot, render)
    self.electricity_cost = 4.25 * WalkerBaseBulletEnv.electricity_cost
    self.stall_torque_cost = 4.25 * WalkerBaseBulletEnv.stall_torque_cost


class HumanoidFlagrunBulletEnv(HumanoidBulletEnv):
  random_yaw = True

  def __init__(self, render=False):
    self.robot = HumanoidFlagrun()
    HumanoidBulletEnv.__init__(self, self.robot, render)

  def create_single_player_scene(self, bullet_client):
    s = HumanoidBulletEnv.create_single_player_scene(self, bullet_client)
    s.zero_at_running_strip_start_line = False
    return s


class HumanoidFlagrunHarderBulletEnv(HumanoidBulletEnv):
  random_lean = True  # can fall on start

  def __init__(self, render=False):
    self.robot = HumanoidFlagrunHarder()
    self.electricity_cost /= 4  # don't care that much about electricity, just stand up!
    HumanoidBulletEnv.__init__(self, self.robot, render)

  def create_single_player_scene(self, bullet_client):
    s = HumanoidBulletEnv.create_single_player_scene(self, bullet_client)
    s.zero_at_running_strip_start_line = False
    return s



import pybullet as p
import time
import math
from datetime import datetime
from datetime import datetime
import pybullet_data

clid = p.connect(p.SHARED_MEMORY)


if (clid < 0):
  p.connect(p.GUI)

p.setPhysicsEngineParameter(enableConeFriction=0)

p.setAdditionalSearchPath(pybullet_data.getDataPath())


p.loadURDF("plane.urdf", [0, 0, -0.3])
husky = p.loadURDF("husky/husky.urdf", [0.290388, 0.329902, -0.310270],
                   [0.002328, -0.000984, 0.996491, 0.083659])
for i in range(p.getNumJoints(husky)):
  print(p.getJointInfo(husky, i))
kukaId = p.loadURDF("kuka_iiwa/model_free_base.urdf", 0.193749, 0.345564, 0.120208, 0.002327,
                    -0.000988, 0.996491, 0.083659)
ob = kukaId
jointPositions = [3.559609, 0.411182, 0.862129, 1.744441, 0.077299, -1.129685, 0.006001]
for jointIndex in range(p.getNumJoints(ob)):
  p.resetJointState(ob, jointIndex, jointPositions[jointIndex])

#put kuka on top of husky

cid = p.createConstraint(husky, -1, kukaId, -1, p.JOINT_FIXED, [0, 0, 0], [0, 0, 0], [0., 0., -.5],
                         [0, 0, 0, 1])

baseorn = p.getQuaternionFromEuler([3.1415, 0, 0.3])
baseorn = [0, 0, 0, 1]
#[0, 0, 0.707, 0.707]

#p.resetBasePositionAndOrientation(kukaId,[0,0,0],baseorn)#[0,0,0,1])
kukaEndEffectorIndex = 6
numJoints = p.getNumJoints(kukaId)
if (numJoints != 7):
  exit()

#lower limits for null space
ll = [-.967, -2, -2.96, 0.19, -2.96, -2.09, -3.05]
#upper limits for null space
ul = [.967, 2, 2.96, 2.29, 2.96, 2.09, 3.05]
#joint ranges for null space
jr = [5.8, 4, 5.8, 4, 5.8, 4, 6]
#restposes for null space
rp = [0, 0, 0, 0.5 * math.pi, 0, -math.pi * 0.5 * 0.66, 0]
#joint damping coefficents
jd = [0.1, 0.1, 0.1, 0.1, 0.1, 0.1, 0.1]

for i in range(numJoints):
  p.resetJointState(kukaId, i, rp[i])

p.setGravity(0, 0, -10)
t = 0.
prevPose = [0, 0, 0]
prevPose1 = [0, 0, 0]
hasPrevPose = 0
useNullSpace = 0

useOrientation = 0
#If we set useSimulation=0, it sets the arm pose to be the IK result directly without using dynamic control.
#This can be used to test the IK result accuracy.
useSimulation = 1
useRealTimeSimulation = 1
p.setRealTimeSimulation(useRealTimeSimulation)
#trailDuration is duration (in seconds) after debug lines will be removed automatically
#use 0 for no-removal
trailDuration = 15
basepos = [0, 0, 0]
ang = 0
ang = 0


def accurateCalculateInverseKinematics(kukaId, endEffectorId, targetPos, threshold, maxIter):
  closeEnough = False
  iter = 0
  dist2 = 1e30
  while (not closeEnough and iter < maxIter):
    jointPoses = p.calculateInverseKinematics(kukaId, kukaEndEffectorIndex, targetPos)
    for i in range(numJoints):
      p.resetJointState(kukaId, i, jointPoses[i])
    ls = p.getLinkState(kukaId, kukaEndEffectorIndex)
    newPos = ls[4]
    diff = [targetPos[0] - newPos[0], targetPos[1] - newPos[1], targetPos[2] - newPos[2]]
    dist2 = (diff[0] * diff[0] + diff[1] * diff[1] + diff[2] * diff[2])
    closeEnough = (dist2 < threshold)
    iter = iter + 1
  #print ("Num iter: "+str(iter) + "threshold: "+str(dist2))
  return jointPoses


wheels = [2, 3, 4, 5]
#(2, b'front_left_wheel', 0, 7, 6, 1, 0.0, 0.0, 0.0, -1.0, 0.0, 0.0, b'front_left_wheel_link')
#(3, b'front_right_wheel', 0, 8, 7, 1, 0.0, 0.0, 0.0, -1.0, 0.0, 0.0, b'front_right_wheel_link')
#(4, b'rear_left_wheel', 0, 9, 8, 1, 0.0, 0.0, 0.0, -1.0, 0.0, 0.0, b'rear_left_wheel_link')
#(5, b'rear_right_wheel', 0, 10, 9, 1, 0.0, 0.0, 0.0, -1.0, 0.0, 0.0, b'rear_right_wheel_link')
wheelVelocities = [0, 0, 0, 0]
wheelDeltasTurn = [1, -1, 1, -1]
wheelDeltasFwd = [1, 1, 1, 1]
while 1:
  keys = p.getKeyboardEvents()
  shift = 0.01
  wheelVelocities = [0, 0, 0, 0]
  speed = 1.0
  for k in keys:
    if ord('s') in keys:
      p.saveWorld("state.py")
    if ord('a') in keys:
      basepos = basepos = [basepos[0], basepos[1] - shift, basepos[2]]
    if ord('d') in keys:
      basepos = basepos = [basepos[0], basepos[1] + shift, basepos[2]]

    if p.B3G_LEFT_ARROW in keys:
      for i in range(len(wheels)):
        wheelVelocities[i] = wheelVelocities[i] - speed * wheelDeltasTurn[i]
    if p.B3G_RIGHT_ARROW in keys:
      for i in range(len(wheels)):
        wheelVelocities[i] = wheelVelocities[i] + speed * wheelDeltasTurn[i]
    if p.B3G_UP_ARROW in keys:
      for i in range(len(wheels)):
        wheelVelocities[i] = wheelVelocities[i] + speed * wheelDeltasFwd[i]
    if p.B3G_DOWN_ARROW in keys:
      for i in range(len(wheels)):
        wheelVelocities[i] = wheelVelocities[i] - speed * wheelDeltasFwd[i]

  baseorn = p.getQuaternionFromEuler([0, 0, ang])
  for i in range(len(wheels)):
    p.setJointMotorControl2(husky,
                            wheels[i],
                            p.VELOCITY_CONTROL,
                            targetVelocity=wheelVelocities[i],
                            force=1000)
  #p.resetBasePositionAndOrientation(kukaId,basepos,baseorn)#[0,0,0,1])
  if (useRealTimeSimulation):
    t = time.time()  #(dt, micro) = datetime.utcnow().strftime('%Y-%m-%d %H:%M:%S.%f').split('.')
    #t = (dt.second/60.)*2.*math.pi
  else:
    t = t + 0.001

  if (useSimulation and useRealTimeSimulation == 0):
    p.stepSimulation()

  for i in range(1):
    #pos = [-0.4,0.2*math.cos(t),0.+0.2*math.sin(t)]
    pos = [0.2 * math.cos(t), 0, 0. + 0.2 * math.sin(t) + 0.7]
    #end effector points down, not up (in case useOrientation==1)
    orn = p.getQuaternionFromEuler([0, -math.pi, 0])

    if (useNullSpace == 1):
      if (useOrientation == 1):
        jointPoses = p.calculateInverseKinematics(kukaId, kukaEndEffectorIndex, pos, orn, ll, ul,
                                                  jr, rp)
      else:
        jointPoses = p.calculateInverseKinematics(kukaId,
                                                  kukaEndEffectorIndex,
                                                  pos,
                                                  lowerLimits=ll,
                                                  upperLimits=ul,
                                                  jointRanges=jr,
                                                  restPoses=rp)
    else:
      if (useOrientation == 1):
        jointPoses = p.calculateInverseKinematics(kukaId,
                                                  kukaEndEffectorIndex,
                                                  pos,
                                                  orn,
                                                  jointDamping=jd)
      else:
        threshold = 0.001
        maxIter = 100
        jointPoses = accurateCalculateInverseKinematics(kukaId, kukaEndEffectorIndex, pos,
                                                        threshold, maxIter)

    if (useSimulation):
      for i in range(numJoints):
        p.setJointMotorControl2(bodyIndex=kukaId,
                                jointIndex=i,
                                controlMode=p.POSITION_CONTROL,
                                targetPosition=jointPoses[i],
                                targetVelocity=0,
                                force=500,
                                positionGain=1,
                                velocityGain=0.1)
    else:
      #reset the joint state (ignoring all dynamics, not recommended to use during simulation)
      for i in range(numJoints):
        p.resetJointState(kukaId, i, jointPoses[i])

  ls = p.getLinkState(kukaId, kukaEndEffectorIndex)
  if (hasPrevPose):
    p.addUserDebugLine(prevPose, pos, [0, 0, 0.3], 1, trailDuration)
    p.addUserDebugLine(prevPose1, ls[4], [1, 0, 0], 1, trailDuration)
  prevPose = pos
  prevPose1 = ls[4]
  hasPrevPose = 1



from pybullet_envs.minitaur.envs.minitaur_alternating_legs_env import MinitaurAlternatingLegsEnv
from pybullet_envs.minitaur.envs.minitaur_ball_gym_env import MinitaurBallGymEnv
#from pybullet_envs.minitaur.envs.minitaur_randomize_terrain_gym_env import MinitaurRandomizeTerrainGymEnv
from pybullet_envs.minitaur.envs.minitaur_reactive_env import MinitaurReactiveEnv
from pybullet_envs.minitaur.envs.minitaur_stand_gym_env import MinitaurStandGymEnv
from pybullet_envs.minitaur.envs.minitaur_trotting_env import MinitaurTrottingEnv
from pybullet_envs.minitaur.envs.minitaur_four_leg_stand_env import MinitaurFourLegStandEnv
from pybullet_envs.minitaur.envs.minitaur_extended_env import MinitaurExtendedEnv, MinitaurReactiveEnv



r"""An example to use simple_ppo_agent.

A galloping example:
blaze run -c opt \
//robotics/reinforcement_learning/minitaur/agents:simple_ppo_agent_example -- \
--logdir=/cns/ij-d/home/jietan/experiment/minitaur_vizier_study_ppo/\
minreact_nonexp_nr_01_186515603_186518344/15/ \
--checkpoint=model.ckpt-14000000

A trotting example:
blaze run -c opt \
//robotics/reinforcement_learning/minitaur/agents:simple_ppo_agent_example -- \
--logdir=/cns/ij-d/home/jietan/experiment/minitaur_vizier_study_ppo/\
mintrot_nonexp_rd_01_186515603_186518344/24/ \
--checkpoint=model.ckpt-14000000

"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import os
import time

import tf.compat.v1 as tf
from pybullet_envs.agents import utility
from pybullet_envs.minitaur.agents import simple_ppo_agent

flags = tf.app.flags
FLAGS = tf.app.flags.FLAGS
flags.DEFINE_string("logdir", None, "The directory that contains checkpoint and config.")
flags.DEFINE_string("checkpoint", None, "The checkpoint file path.")
flags.DEFINE_string("log_path", None, "The output path to write log.")


def main(argv):
  del argv  # Unused.
  config = utility.load_config(FLAGS.logdir)
  policy_layers = config.policy_layers
  value_layers = config.value_layers
  env = config.env(render=True, log_path=FLAGS.log_path, env_randomizer=None)
  network = config.network

  with tf.Session() as sess:
    agent = simple_ppo_agent.SimplePPOPolicy(sess,
                                             env,
                                             network,
                                             policy_layers=policy_layers,
                                             value_layers=value_layers,
                                             checkpoint=os.path.join(FLAGS.logdir,
                                                                     FLAGS.checkpoint))

    sum_reward = 0
    observation = env.reset()
    while True:
      action = agent.get_action([observation])
      observation, reward, done, _ = env.step(action[0])
      # This sleep is to prevent serial communication error on the real robot.
      time.sleep(0.002)
      sum_reward += reward
      if done:
        break
    tf.logging.info("reward: {}".format(sum_reward))


if __name__ == "__main__":
  tf.app.run(main)



"""A Raibert style controller for Minitaur."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import math
import attr
import gin
import numpy as np

from pybullet_envs.minitaur.agents.baseline_controller import minitaur_raibert_controller_utils
from pybullet_envs.minitaur.envs.utilities import minitaur_pose_utils

LEFT_FRONT_LEG_ID = 0
LEFT_HIND_LEG_ID = 1
RIGHT_FRONT_LEG_ID = 2
RIGHT_HIND_LEG_ID = 3

DIAGONAL_LEG_PAIR_1 = (LEFT_FRONT_LEG_ID, RIGHT_HIND_LEG_ID)
DIAGONAL_LEG_PAIR_2 = (LEFT_HIND_LEG_ID, RIGHT_FRONT_LEG_ID)

LEFT_LEG_IDS = (LEFT_FRONT_LEG_ID, LEFT_HIND_LEG_ID)
RIGHT_LEG_IDS = (RIGHT_FRONT_LEG_ID, RIGHT_HIND_LEG_ID)

# The max horizontal foot offset (in meters) for turning.
_FOOT_HORIZONTAL_OFFSET_FOR_TURNING = 0.1

_STANCE_TG_PHASE_COMPRESSION = 1.5
_STANCE_TG_DELTA_EXTENSION = 0.2
_STANCE_HORIZONTAL_SCALING_FACTOR = 1.2

_DEFAULT_SWING_SPEED_GAIN = 0.015

_DEFAULT_SWING_FOOT_CLEARANCE = 0.005

_PITCH_SWING_FEEDBACK_SCALING_FACTOR = 1.2


# A POD container to describe the controller's high level behavior.
@attr.s
class BehaviorParameters(object):
  """Highlevel parameters for Raibert style controller."""
  stance_duration = attr.ib(type=float, default=0.25)
  desired_forward_speed = attr.ib(type=float, default=0.)
  desired_turning_speed = attr.ib(type=float, default=0.)
  standing_height = attr.ib(type=float, default=0.2)
  desired_incline_angle = attr.ib(type=float, default=0.)


def generate_default_swing_trajectory(phase, init_pose, end_pose):
  """A swing trajectory generator.

  Args:
    phase: Float. Between [0, 1].
    init_pose: A tuple. The leg pose (swing, extension) at phase == 0 the
      beginning of swing.
    end_pose: A tuple. The leg pose at phase == 1 the end of swing.

  Returns:
    The desired leg pose for the current phase.
  """
  # Try phase compression
  normalized_phase = math.sqrt(min(phase * 1.5, 1))

  # For swing, we use a linear interpolation:
  swing = (end_pose[0] - init_pose[0]) * normalized_phase + init_pose[0]

  # For extension, we can fit a second order polynomial:
  min_ext = (init_pose[1] + end_pose[1]) / 2 - 0.8
  min_ext = max(min_ext, 0.5)

  # The phase value at which the extension reaches the minimal value min_ext.
  # phi is small, the swing leg will try to lift higher in the first half of
  # swing.
  phi = 0.1

  # We convert the extension back into the cartesion space. In this way we can
  # guarantee a lift-up trajectory. The ankle to hip distance is easier to
  # compute than a full forward-kinematics.
  min_ankle_dist = minitaur_raibert_controller_utils.extension_to_ankle_dist(
      min_ext)
  init_ankle_dist = minitaur_raibert_controller_utils.extension_to_ankle_dist(
      init_pose[1])
  end_ankle_dist = minitaur_raibert_controller_utils.extension_to_ankle_dist(
      end_pose[1])

  # The polynomial is: a * phi^2 + b * phi + c
  delta_1 = min_ankle_dist - init_ankle_dist
  delta_2 = end_ankle_dist - init_ankle_dist
  delta_p = phi * phi - phi

  a = (delta_1 - phi * delta_2) / delta_p

  b = (phi * phi * delta_2 - delta_1) / delta_p

  ankle_dist = (
      a * normalized_phase * normalized_phase + b * normalized_phase +
      init_ankle_dist)

  l1 = minitaur_raibert_controller_utils.UPPER_LEG_LEN
  l2 = minitaur_raibert_controller_utils.LOWER_SHORT_LEG_LEN

  ankle_dist = min(max(ankle_dist, l2 - l1 + 0.01), l2 + l1 - 0.01)

  extension = minitaur_raibert_controller_utils.ankle_dist_to_extension(
      ankle_dist)

  return (swing, extension)


@gin.configurable
def generate_default_stance_trajectory(phase,
                                       init_pose,
                                       end_pose,
                                       use_constant_extension=False):
  """A stance strajectory generator.

  Args:
    phase: Float. Between [0, 1].
    init_pose: A tuple. The leg pose (swing, extension) at phase == 0, i.e. the
      beginning of stance.
    end_pose: A tuple. The leg pose at phase == 1, i.e. the end of stance.
    use_constant_extension: Boolean. Whether or not to fix the extension during
      stance.

  Returns:
    The desired leg pose for the current phase.
  """
  normalized_phase = min(_STANCE_TG_PHASE_COMPRESSION * math.sqrt(phase), 1)
  swing = (end_pose[0] - init_pose[0]) * normalized_phase + init_pose[0]

  # The extension evolves nonlinearly according to the parabola equation.
  if use_constant_extension:
    extension = end_pose[1]
  else:
    extension = end_pose[1] - 4 * _STANCE_TG_DELTA_EXTENSION * (
        normalized_phase**2 - normalized_phase)
  return (swing, extension)


def get_stance_foot_offset_for_turning(leg_id, steering_signal):
  """Modify the stance foot position to achieve turning.

  The strategy works for trotting gaits.

  Args:
    leg_id: Integer. The leg index.
    steering_signal: Float. The desired turning signal in [-1, 1]. Because we
      don't have an accurate mapping from angular velocity to the foot offset,
      the steering signal should be treated as a reference and only its relative
      magnitude matters.

  Returns:
    The stance foot's horizontal offset.

  """
  clipped_steering_signal = np.clip(steering_signal, -1, 1)

  if leg_id in LEFT_LEG_IDS:
    return _FOOT_HORIZONTAL_OFFSET_FOR_TURNING * clipped_steering_signal
  else:
    return -(_FOOT_HORIZONTAL_OFFSET_FOR_TURNING * clipped_steering_signal)


def get_leg_swing_offset_for_pitching(body_pitch, desired_incline_angle):
  """Get the leg swing zero point when the body is tilted.

  For example, when climbing up or down stairs/slopes, the robot body will tilt
  up or down. By compensating the body pitch, the leg's trajectory will be
  centered around the vertical direction (not perpendicular to the surface).
  This helps the robot to generate thrust when going upwards, and braking when
  going downwards.

  Args:
    body_pitch: Float. Current body pitch angle.
    desired_incline_angle: Float. The desired body pitch angle.

  Returns:
    The stance and swing leg swing offset.

  """
  kp = 0.2
  return -((1 - kp) * body_pitch + kp * desired_incline_angle)


@gin.configurable
class RaibertSwingLegController(object):
  """The swing leg controller."""

  def __init__(self,
               speed_gain=_DEFAULT_SWING_SPEED_GAIN,
               foot_clearance=_DEFAULT_SWING_FOOT_CLEARANCE,
               leg_trajectory_generator=generate_default_swing_trajectory):
    """Initializes the controller.

    Args:
      speed_gain: Float. The speed feedback gain to modulate the target foot
        position.
      foot_clearance: Float. The foot clearance (at the end of the swing phase)
        in meters.
      leg_trajectory_generator: A trajectory generator function.
    """
    self._speed_gain = speed_gain
    self._foot_clearance = foot_clearance
    self._leg_trajectory_generator = leg_trajectory_generator

  def get_action(self, raibert_controller):
    """Get the swing legs' desired pose."""
    current_speed = raibert_controller.estimate_base_velocity()
    phase = raibert_controller.get_phase()
    rpy = raibert_controller.robot.base_roll_pitch_yaw

    leg_pose_set = {}
    for i in raibert_controller.swing_set:
      # Target foot horizontal position is calculated according to Raibert's
      # original formula in "Legged robots that balance".
      target_foot_horizontal_position = (
          raibert_controller.behavior_parameters.stance_duration / 2 *
          current_speed + self._speed_gain *
          (current_speed -
           raibert_controller.behavior_parameters.desired_forward_speed))

      # 1) Convert the target foot position to leg pose space.
      # Lift the foot a little bit.
      target_foot_vertical_position = -(
          raibert_controller.behavior_parameters.standing_height -
          self._foot_clearance)
      target_foot_position = (target_foot_horizontal_position,
                              target_foot_vertical_position)
      target_leg_pose = minitaur_raibert_controller_utils.foot_position_to_leg_pose(
          target_foot_position)

      # 2) Generates the curve from the swing start leg pose to the target leg
      # pose and find the next leg pose on the curve based on current swing
      # phase.

      desired_leg_pose = self._leg_trajectory_generator(
          phase, raibert_controller.swing_start_leg_pose, target_leg_pose)

      swing_offset = get_leg_swing_offset_for_pitching(
          rpy[1], raibert_controller.behavior_parameters.desired_incline_angle)

      leg_pose_set[i] = (desired_leg_pose[0] + swing_offset,
                         desired_leg_pose[1])

    return leg_pose_set


@gin.configurable
class RaibertStanceLegController(object):
  """The controller that modulates the behavior of the stance legs."""

  def __init__(self,
               speed_gain=0.1,
               leg_trajectory_generator=generate_default_stance_trajectory):
    """Initializes the controller.

    Args:
      speed_gain: Float. The speed feedback gain to modulate the target stance
        foot position.
      leg_trajectory_generator: A trajectory generator function. Generates the
        desired leg pose during the stance phase.
    """
    self._speed_gain = speed_gain
    self._leg_trajectory_generator = leg_trajectory_generator

  def get_action(self, raibert_controller):
    """Get the desired leg pose for the stance legs."""

    phase = raibert_controller.get_phase()
    current_speed = raibert_controller.estimate_base_velocity()
    rpy = raibert_controller.robot.base_roll_pitch_yaw

    leg_pose_set = {}
    for i in raibert_controller.stance_set:
      desired_forward_speed = (
          raibert_controller.behavior_parameters.desired_forward_speed)

      target_foot_horizontal_position = -_STANCE_HORIZONTAL_SCALING_FACTOR * (
          raibert_controller.behavior_parameters.stance_duration / 2 *
          current_speed - self._speed_gain *
          (current_speed - desired_forward_speed))

      target_foot_horizontal_position += get_stance_foot_offset_for_turning(
          i, raibert_controller.behavior_parameters.desired_turning_speed)

      target_foot_position = (
          target_foot_horizontal_position,
          -raibert_controller.behavior_parameters.standing_height)
      target_leg_pose = minitaur_raibert_controller_utils.foot_position_to_leg_pose(
          target_foot_position)

      desired_leg_pose = (
          self._leg_trajectory_generator(
              phase, raibert_controller.stance_start_leg_pose, target_leg_pose))

      swing_offset = _PITCH_SWING_FEEDBACK_SCALING_FACTOR * get_leg_swing_offset_for_pitching(
          rpy[1], raibert_controller.behavior_parameters.desired_incline_angle)

      leg_pose_set[i] = (desired_leg_pose[0] + swing_offset,
                         desired_leg_pose[1])

    return leg_pose_set


@gin.configurable
class MinitaurRaibertController(object):
  """A Raibert style controller for trotting gait."""

  def __init__(self,
               robot,
               behavior_parameters=BehaviorParameters(),
               swing_leg_controller=RaibertSwingLegController(),
               stance_leg_controller=RaibertStanceLegController(),
               pose_feedback_controller=None):
    self._time = 0
    self._robot = robot
    self.behavior_parameters = behavior_parameters

    self._last_recorded_speed = 0

    self._swing_leg_controller = swing_leg_controller
    self._stance_leg_controller = stance_leg_controller
    self._pose_feeback_controller = pose_feedback_controller

    # The leg order is FL, RL, FR, RR -> [0, 1, 2, 3]
    self._swing_set = DIAGONAL_LEG_PAIR_1
    self._stance_set = DIAGONAL_LEG_PAIR_2

    # Compute the initial leg pose.
    self._swing_start_leg_pose = self.get_swing_leg_pose()
    self._stance_start_leg_pose = self.get_stance_leg_pose()

  @property
  def robot(self):
    return self._robot

  @property
  def swing_set(self):
    return self._swing_set

  @property
  def stance_set(self):
    return self._stance_set

  @property
  def swing_start_leg_pose(self):
    return self._swing_start_leg_pose

  @property
  def stance_start_leg_pose(self):
    return self._stance_start_leg_pose

  def _get_average_leg_pose(self, leg_indices):
    """Get the average leg pose."""
    current_leg_pose = minitaur_pose_utils.motor_angles_to_leg_pose(
        self._robot.motor_angles)

    # extract the swing leg pose from the current_leg_pose
    leg_pose = []
    for index in leg_indices:
      leg_pose.append([
          current_leg_pose[index],
          current_leg_pose[index + minitaur_pose_utils.NUM_LEGS]
      ])

    leg_pose = np.array(leg_pose)
    return np.mean(leg_pose, axis=0)

  def get_swing_leg_pose(self):
    """Get the current swing legs' average pose."""
    return self._get_average_leg_pose(self._swing_set)

  def get_stance_leg_pose(self):
    """Get the current stance legs' average pose."""
    return self._get_average_leg_pose(self._stance_set)

  def get_phase(self):
    """Compute the current stance/swing phase."""
    return math.fmod(self._time, self.behavior_parameters.stance_duration
                    ) / self.behavior_parameters.stance_duration

  def _get_new_swing_stance_set(self):
    """Switch the set of swing/stance legs based on timing."""
    swing_stance_phase = math.fmod(self._time,
                                   2 * self.behavior_parameters.stance_duration)
    if swing_stance_phase < self.behavior_parameters.stance_duration:
      return (DIAGONAL_LEG_PAIR_1, DIAGONAL_LEG_PAIR_2)
    return (DIAGONAL_LEG_PAIR_2, DIAGONAL_LEG_PAIR_1)

  def update(self, t):
    """Update the internal status of the controller.

    Args:
      t: Float. The current time after reset in seconds.
    """
    self._time = t
    new_swing_set, new_stance_set = self._get_new_swing_stance_set()

    # If there is a stance/swing switch.
    if new_swing_set[0] is not self._swing_set[0]:
      self._swing_set = new_swing_set
      self._stance_set = new_stance_set

      # Also records the starting pose.
      self._swing_start_leg_pose = self.get_swing_leg_pose()
      self._stance_start_leg_pose = self.get_stance_leg_pose()

  def estimate_base_velocity(self):
    """Estimate the current CoM speed."""
    # It is best to use a sensor fusion approach.
    stance_leg_pose = self.get_stance_leg_pose()

    delta_sw = stance_leg_pose[0] - self._stance_start_leg_pose[0]

    x, y = minitaur_raibert_controller_utils.leg_pose_to_foot_position(
        stance_leg_pose)
    toe_hip_len = math.sqrt(x**2 + y**2)
    horizontal_dist = toe_hip_len * delta_sw
    phase = self.get_phase()
    speed = self._last_recorded_speed
    if phase > 0.1:
      speed = horizontal_dist / (
          self.behavior_parameters.stance_duration * phase)
      self._last_recorded_speed = speed
    return speed

  def get_swing_leg_action(self):
    return self._swing_leg_controller.get_action(self)

  def get_stance_leg_action(self):
    return self._stance_leg_controller.get_action(self)

  def get_action(self):
    """Gets the desired motor angles."""
    leg_pose = [0] * minitaur_pose_utils.NUM_MOTORS
    swing_leg_pose = self.get_swing_leg_action()
    for i in self._swing_set:
      leg_pose[i] = swing_leg_pose[i][0]
      leg_pose[i + minitaur_pose_utils.NUM_LEGS] = swing_leg_pose[i][1]

    stance_leg_pose = self.get_stance_leg_action()
    for i in self._stance_set:
      leg_pose[i] = stance_leg_pose[i][0]
      leg_pose[i + minitaur_pose_utils.NUM_LEGS] = stance_leg_pose[i][1]

    return minitaur_pose_utils.leg_pose_to_motor_angles(leg_pose)



import pybullet as p
import pybullet_data as pd

import time

p.connect(p.GUI)
p.setAdditionalSearchPath(pd.getDataPath())

p.setPhysicsEngineParameter(enableConeFriction=0)
plane = p.loadURDF("plane.urdf")
p.setGravity(0, 0, -9.8)
p.setTimeStep(1. / 500)
#p.setDefaultContactERP(0)
#urdfFlags = p.URDF_USE_SELF_COLLISION+p.URDF_USE_SELF_COLLISION_EXCLUDE_ALL_PARENTS
urdfFlags = p.URDF_USE_SELF_COLLISION
quadruped = p.loadURDF("laikago/laikago_toes.urdf", [0, 0, .5], [0, 0.5, 0.5, 0],
                       flags=urdfFlags,
                       useFixedBase=False)

#enable collision between lower legs

for j in range(p.getNumJoints(quadruped)):
  print(p.getJointInfo(quadruped, j))

#2,5,8 and 11 are the lower legs
lower_legs = [2, 5, 8, 11]
for l0 in lower_legs:
  for l1 in lower_legs:
    if (l1 > l0):
      enableCollision = 1
      print("collision for pair", l0, l1,
            p.getJointInfo(quadruped, l0)[12],
            p.getJointInfo(quadruped, l1)[12], "enabled=", enableCollision)
      p.setCollisionFilterPair(quadruped, quadruped, 2, 5, enableCollision)

jointIds = []
paramIds = []
jointOffsets = []
jointDirections = [-1, 1, 1, 1, 1, 1, -1, 1, 1, 1, 1, 1]
jointAngles = [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0]

for i in range(4):
  jointOffsets.append(0)
  jointOffsets.append(-0.7)
  jointOffsets.append(0.7)

maxForceId = p.addUserDebugParameter("maxForce", 0, 100, 20)

for j in range(p.getNumJoints(quadruped)):
  p.changeDynamics(quadruped, j, linearDamping=0, angularDamping=0)
  info = p.getJointInfo(quadruped, j)
  #print(info)
  jointName = info[1]
  jointType = info[2]
  if (jointType == p.JOINT_PRISMATIC or jointType == p.JOINT_REVOLUTE):
    jointIds.append(j)

p.getCameraImage(480, 320)
p.setRealTimeSimulation(0)

joints = []

with open(pd.getDataPath() + "/laikago/data1.txt", "r") as filestream:
  for line in filestream:
    print("line=", line)
    maxForce = p.readUserDebugParameter(maxForceId)
    currentline = line.split(",")
    #print (currentline)
    #print("-----")
    frame = currentline[0]
    t = currentline[1]
    #print("frame[",frame,"]")
    joints = currentline[2:14]
    #print("joints=",joints)
    for j in range(12):
      targetPos = float(joints[j])
      p.setJointMotorControl2(quadruped,
                              jointIds[j],
                              p.POSITION_CONTROL,
                              jointDirections[j] * targetPos + jointOffsets[j],
                              force=maxForce)
    p.stepSimulation()
    for lower_leg in lower_legs:
      #print("points for ", quadruped, " link: ", lower_leg)
      pts = p.getContactPoints(quadruped, -1, lower_leg)
      #print("num points=",len(pts))
      #for pt in pts:
      # print(pt[9])
    time.sleep(1. / 500.)

index = 0
for j in range(p.getNumJoints(quadruped)):
  p.changeDynamics(quadruped, j, linearDamping=0, angularDamping=0)
  info = p.getJointInfo(quadruped, j)
  js = p.getJointState(quadruped, j)
  #print(info)
  jointName = info[1]
  jointType = info[2]
  if (jointType == p.JOINT_PRISMATIC or jointType == p.JOINT_REVOLUTE):
    paramIds.append(p.addUserDebugParameter(jointName.decode("utf-8"), -4, 4,
                                (js[0] - jointOffsets[index]) / jointDirections[index]))
    index = index+1
                       

p.setRealTimeSimulation(1)

while (1):

  for i in range(len(paramIds)):
    c = paramIds[i]
    targetPos = p.readUserDebugParameter(c)
    maxForce = p.readUserDebugParameter(maxForceId)
    p.setJointMotorControl2(quadruped,
                            jointIds[i],
                            p.POSITION_CONTROL,
                            jointDirections[i] * targetPos + jointOffsets[i],
                            force=maxForce)



# Copyright 2017 The TensorFlow Agents Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
r"""Script to train a batch reinforcement learning algorithm.

Command line:

  python3 -m agents.scripts.train --logdir=/path/to/logdir --config=pendulum
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import datetime
import os

import gym
try:
  import tensorflow.compat.v1 as tf
except Exception:
  import tensorflow as tf

from . import tools
from . import configs
from . import utility


def _create_environment(config):
  """Constructor for an instance of the environment.

  Args:
    config: Object providing configurations via attributes.

  Returns:
    Wrapped OpenAI Gym environment.
  """
  if isinstance(config.env, str):
    env = gym.make(config.env)
  else:
    env = config.env()
  if config.max_length:
    env = tools.wrappers.LimitDuration(env, config.max_length)
  env = tools.wrappers.RangeNormalize(env)
  env = tools.wrappers.ClipAction(env)
  env = tools.wrappers.ConvertTo32Bit(env)
  return env


def _define_loop(graph, logdir, train_steps, eval_steps):
  """Create and configure a training loop with training and evaluation phases.

  Args:
    graph: Object providing graph elements via attributes.
    logdir: Log directory for storing checkpoints and summaries.
    train_steps: Number of training steps per epoch.
    eval_steps: Number of evaluation steps per epoch.

  Returns:
    Loop object.
  """
  loop = tools.Loop(logdir, graph.step, graph.should_log, graph.do_report, graph.force_reset)
  loop.add_phase('train',
                 graph.done,
                 graph.score,
                 graph.summary,
                 train_steps,
                 report_every=train_steps,
                 log_every=train_steps // 2,
                 checkpoint_every=None,
                 feed={graph.is_training: True})
  loop.add_phase('eval',
                 graph.done,
                 graph.score,
                 graph.summary,
                 eval_steps,
                 report_every=eval_steps,
                 log_every=eval_steps // 2,
                 checkpoint_every=10 * eval_steps,
                 feed={graph.is_training: False})
  return loop


def train(config, env_processes):
  """Training and evaluation entry point yielding scores.

  Resolves some configuration attributes, creates environments, graph, and
  training loop. By default, assigns all operations to the CPU.

  Args:
    config: Object providing configurations via attributes.
    env_processes: Whether to step environments in separate processes.

  Yields:
    Evaluation scores.
  """
  tf.reset_default_graph()
  if config.update_every % config.num_agents:
    tf.logging.warn('Number of agents should divide episodes per update.')
  with tf.device('/cpu:0'):
    batch_env = utility.define_batch_env(lambda: _create_environment(config), config.num_agents,
                                         env_processes)
    graph = utility.define_simulation_graph(batch_env, config.algorithm, config)
    loop = _define_loop(graph, config.logdir, config.update_every * config.max_length,
                        config.eval_episodes * config.max_length)
    total_steps = int(config.steps / config.update_every *
                      (config.update_every + config.eval_episodes))
  # Exclude episode related variables since the Python state of environments is
  # not checkpointed and thus new episodes start after resuming.
  saver = utility.define_saver(exclude=(r'.*_temporary/.*',))
  sess_config = tf.ConfigProto(allow_soft_placement=True)
  sess_config.gpu_options.allow_growth = True
  with tf.Session(config=sess_config) as sess:
    utility.initialize_variables(sess, saver, config.logdir)
    for score in loop.run(sess, saver, total_steps):
      yield score
  batch_env.close()


def main(_):
  """Create or load configuration and launch the trainer."""
  utility.set_up_logging()
  if not FLAGS.config:
    raise KeyError('You must specify a configuration.')
  logdir = FLAGS.logdir and os.path.expanduser(
      os.path.join(FLAGS.logdir, '{}-{}'.format(FLAGS.timestamp, FLAGS.config)))
  try:
    config = utility.load_config(logdir)
  except IOError:
    config = tools.AttrDict(getattr(configs, FLAGS.config)())
    config = utility.save_config(config, logdir)
  for score in train(config, FLAGS.env_processes):
    tf.logging.info('Score {}.'.format(score))


if __name__ == '__main__':
  FLAGS = tf.app.flags.FLAGS
  tf.app.flags.DEFINE_string('logdir', None, 'Base directory to store logs.')
  tf.app.flags.DEFINE_string('timestamp',
                             datetime.datetime.now().strftime('%Y%m%dT%H%M%S'),
                             'Sub directory to store logs.')
  tf.app.flags.DEFINE_string('config', None, 'Configuration to execute.')
  tf.app.flags.DEFINE_boolean('env_processes', True,
                              'Step environments in separate processes to circumvent the GIL.')
  tf.app.run()
